"""Local model access via the llama.cpp engine.

- LlamaCppModel: Use local GGUF format models via llama.cpp engine.
- LlamaCppTokenizer: Tokenizer for llama.cpp loaded GGUF models.
"""


from typing import Any, Optional, Union

import sys, os, json, ctypes
from copy import copy

import logging
logger = logging.getLogger(__name__)


from .gen import (
    GenConf
)

from .thread import (
    Thread
)

from .model import (
    FormattedTextModel,
    Tokenizer
)

from .json_schema import JSchemaConf

from .json_grammar import (
    gbnf_from_json_schema,
    JSON_GBNF
)

try:
    from llama_cpp import (
        Llama,
        llama_cpp,
        llama_token_get_text,
        llama_grammar
    )
    has_llama_cpp = True
except ImportError:
    has_llama_cpp = False






class LlamaCppModel(FormattedTextModel):
    """Use local GGUF format models via llama.cpp engine.
    
    Supports grammar-constrained JSON output following a JSON schema.

    Attributes:
        ctx_len: Maximum context length, shared for input + output.
        desc: Model information.
    """

    _llama: Llama
    """LlamaCpp instance"""

    def __init__(self,
                 path: str,

                 format: Optional[str] = None,                 
                 format_search_order: list[str] = ["name","meta_template", "formats_json"],

                 *,

                 # common base model args
                 genconf: Optional[GenConf] = None,
                 schemaconf: Optional[JSchemaConf] = None,
                 tokenizer: Optional[Tokenizer] = None,
                 ctx_len: int = 2048,

                 # important LlamaCpp-specific args
                 n_gpu_layers: int = -1,
                 main_gpu: int = 0,
                 n_batch: int = 512,
                 seed: int = 4294967295,
                 verbose: bool = False,

                 # other LlamaCpp-specific args
                 **llamacpp_kwargs
                 ):
        """
        Args:
            path: File path to the GGUF file.
            format: Chat template format to use with model. Leave as None for auto-detection.
            format_search_order: Search order for auto-detecting format, "name" searches in the filename, "meta_template" looks in the model's metadata. Defaults to ["name","meta_template"].
            genconf: Default generation configuration, which can be used in gen() and related. Defaults to None.
            tokenizer: An external initialized tokenizer to use instead of the created from the GGUF file. Defaults to None.
            ctx_len: Maximum context length to be used (shared for input and output). Defaults to 2048.
            n_gpu_layers: Number of model layers to run in a GPU. Defaults to -1 for all.
            main_gpu: Index of the GPU to use. Defaults to 0.
            n_batch: Prompt processing batch size. Defaults to 512.
            seed: Random number generation seed, for non zero temperature inference. Defaults to 4294967295.
            verbose: Emit (very) verbose output. Defaults to False.

        Raises:
            ImportError: If llama-cpp-python is not installed.
            ValueError: If ctx_len is 0 or larger than the values supported by model.
        """

        self._llama = None # type: ignore[assignment]
        self.tokenizer = None # type: ignore[assignment]

        if not has_llama_cpp:
            raise ImportError("Please install llama-cpp-python by running: pip install llama-cpp-python")

        if ctx_len == 0:
            raise ValueError("LlamaCppModel doesn't support ctx_len=0")
        
        super().__init__(True,
                         genconf,
                         schemaconf,
                         tokenizer
                         )

        # update kwargs from important args
        llamacpp_kwargs.update(n_ctx=ctx_len,
                               n_batch=n_batch,
                               n_gpu_layers=n_gpu_layers,
                               main_gpu=main_gpu,
                               seed=seed,
                               verbose=verbose
                               )
        
        logger.debug(f"Creating inner Llama with model_path='{path}', llamacpp_kwargs={llamacpp_kwargs}")

        with normalize_notebook_stdout_stderr(not verbose):
            self._llama = Llama(model_path=path, **llamacpp_kwargs)

        self._model_path = path
        
        # correct super __init__ values
        self._ctx_len = self._llama.n_ctx()
        
        n_ctx_train = self._llama._model.n_ctx_train()        
        if self.ctx_len > n_ctx_train:
            raise ValueError(f"ctx_len ({self.ctx_len}) is greater than n_ctx_train ({n_ctx_train})")

        
        if self.tokenizer is None:
            self.tokenizer = LlamaCppTokenizer(self._llama)

        try:
            self.init_format(format,
                             format_search_order,
                             {"name": os.path.basename(self._model_path),
                              "path": self._model_path,
                              "meta_template_name": "tokenizer.chat_template"}
                             )
        except Exception as e:
            del self.tokenizer
            del self._llama
            raise e
            
        
   
    

    def __del__(self):
        if hasattr(self, "tokenizer"):
            del self.tokenizer
        if hasattr(self, "_llama"):
            if self._llama:
                del self._llama

    
    
    
    def _gen_text(self,
                  text: str,
                  genconf: GenConf) -> tuple[str,str]:
        """Generate from formatted text.

        Args:
            text: Formatted text (from input Thread).
            genconf: Model generation configuration.

        Returns:
            Tuple of strings: generated_text, finish_reason.
        """

        token_ids = self.tokenizer.encode(text)

        if genconf is None: 
            genconf = self.genconf

        if genconf.max_tokens == 0:
            genconf = genconf(max_tokens=self.ctx_len - len(token_ids))
            
        elif len(token_ids) + genconf.max_tokens > self.ctx_len:
            logger.warn(f"Token length + genconf.max_tokens ({len(token_ids) + genconf.max_tokens}) is greater than model's context window length ({self.ctx_len})")

        # prepare llamaCpp args:
        genconf_kwargs = genconf.as_dict()
        
        format = genconf_kwargs.pop("format")
        if format == "json":
            if genconf_kwargs["json_schema"] is None:
                grammar = llama_grammar.LlamaGrammar.from_string(JSON_GBNF, 
                                                                 logger.getEffectiveLevel() == logging.DEBUG)
                    
            else: # translate json_schema to a llama grammar
                jsg = gbnf_from_json_schema(genconf_kwargs["json_schema"])
                logger.debug(f"JSON schema GBNF grammar:\n{jsg}")
                
                grammar = llama_grammar.LlamaGrammar.from_string(jsg, 
                                                                 logger.getEffectiveLevel() == logging.DEBUG)

            genconf_kwargs["grammar"] = grammar
            
        genconf_kwargs.pop("json_schema")

        logger.debug(f"LlamaCpp args: {genconf_kwargs}")
        
        # Llamacpp.create_completion() never returns special tokens because it uses its own detokenize()
        response = self._llama.create_completion(token_ids,
                                                 echo=False,
                                                 stream=False,
                                                 **genconf_kwargs)
        logger.debug(f"LlamaCpp response: {response}")

        choice = response["choices"][0] # type: ignore[index]
        return choice["text"], choice["finish_reason"] # type: ignore[return-value]


        
    @classmethod
    def provider_version(_) -> str:
        """Provider library version: provider x.y.z
        Ex. llama-cpp-python 0.2.44
        """
        try:        
            import llama_cpp
            ver = llama_cpp.__version__
        except Exception:
            raise ImportError("Please install llama-cpp-python by running: pip install llama-cpp-python")
            
        return f"llama-cpp-python {ver}"

    
    @property
    def desc(self) -> str:
        """Model description."""
        return f"LlamaCppModel: {self._model_path} - '{self._llama._model.desc()}'"

    
    @property
    def n_embd(self) -> int:
        """Embedding size of model."""
        return self._llama.n_embd()

    @property
    def n_params(self) -> int:
        """Total number of model parameters."""
        return self._llama._model.n_params()



    def get_metadata(self):
        """Returns model metadata."""
        out = {}
        buf = bytes(16 * 1024)
        lmodel = self._llama.model
        count = llama_cpp.llama_model_meta_count(lmodel)
        for i in range(count):
            res = llama_cpp.llama_model_meta_key_by_index(lmodel, i, buf,len(buf))
            if res >= 0:
                key = buf[:res].decode('utf-8')
                res = llama_cpp.llama_model_meta_val_str_by_index(lmodel, i, buf,len(buf))
                if res >= 0:
                    value = buf[:res].decode('utf-8')
                    out[key] = value
        return out


   
    
    


class LlamaCppTokenizer(Tokenizer):
    """Tokenizer for llama.cpp loaded GGUF models."""

    def __init__(self, 
                 llama: Llama, 
                 reg_flags: Optional[str] = None):
        self._llama = llama

        self.vocab_size = self._llama.n_vocab()

        self.bos_token_id = self._llama.token_bos()
        self.bos_token = llama_token_get_text(self._llama.model, self.bos_token_id).decode("utf-8")

        self.eos_token_id = self._llama.token_eos()
        self.eos_token = llama_token_get_text(self._llama.model, self.eos_token_id).decode("utf-8")

        self.pad_token_id = None
        self.pad_token = None

        self.unk_token_id = None # ? fill by taking a look at id 0?
        self.unk_token = None

        # workaround for https://github.com/ggerganov/llama.cpp/issues/4772
        self._workaround1 = reg_flags is not None and "llamacpp1" in reg_flags
            

    
    def encode(self, 
               text: str) -> list[int]:
        """Encode text into model tokens. Inverse of Decode().

        Args:
            text: Text to be encoded.

        Returns:
            A list of ints with the encoded tokens.
        """

        if self._workaround1:
            # append a space after each bos and eos, so that llama's tokenizer matches HF
            def space_post(text, s):
                out = ""
                while (index := text.find(s)) != -1:
                    after = index + len(s)
                    out += text[:after]
                    if text[after] != ' ':
                        out += ' '
                    text = text[after:]
                        
                out += text
                return out

            text = space_post(text, self.bos_token)
            text = space_post(text, self.eos_token)
            # print(text)
        
        # str -> bytes
        btext = text.encode("utf-8", errors="ignore")

        return self._llama.tokenize(btext, add_bos=False, special=True)



    def decode(self,
               token_ids: list[int],
               skip_special: bool = True) -> str:
        """Decode model tokens to text. Inverse of Encode().

        Using instead of llama-cpp-python's to fix error: remove first character after a bos only if it's a space.

        Args:
            token_ids: List of model tokens.
            skip_special: Don't decode special tokens like bos and eos. Defaults to True.

        Returns:
            Decoded text.
        """

        if not len(token_ids):
            return ""
        
        output = b""
        size = 32
        buffer = (ctypes.c_char * size)()

        if not skip_special:
            special_toks = {self.bos_token_id: self.bos_token.encode("utf-8"), # type: ignore[union-attr]
                            self.eos_token_id: self.eos_token.encode("utf-8")} # type: ignore[union-attr]

            for token in token_ids:
                if token == self.bos_token_id:
                    output += special_toks[token]
                elif token == self.eos_token_id:
                    output += special_toks[token]
                else:
                    n = llama_cpp.llama_token_to_piece(
                        self._llama.model, llama_cpp.llama_token(token), buffer, size
                    )
                    output += bytes(buffer[:n]) # type: ignore[arg-type]

        else: # skip special
            for token in token_ids:
                if token != self.bos_token_id and token != self.eos_token_id:
                    n = llama_cpp.llama_token_to_piece(
                        self._llama.model, llama_cpp.llama_token(token), buffer, size
                    )
                    output += bytes(buffer[:n]) # type: ignore[arg-type]
            

        # "User code is responsible for removing the leading whitespace of the first non-BOS token when decoding multiple tokens."
        if (# token_ids[0] != self.bos_token_id and # we also try cutting if first is bos to approximate HF tokenizer
           len(output) and output[0] <= 32 # 32 = ord(' ')
           ):
            output = output[1:]
                
        return output.decode("utf-8", errors="ignore")
    
        




class normalize_notebook_stdout_stderr():
    '''
    Based on guidance._utils.py (MIT License):
        Remaps stdout and stderr back to their normal selves from what ipykernel did to them.
        Based on: https://github.com/ipython/ipykernel/issues/795
    '''
    def __init__(self, 
                 enabled: bool = True):
        self._enabled = enabled

    def __enter__(self):
        if self._enabled:
            normal_stdout = sys.__stdout__.fileno()
            self.restore_stdout = None
            if getattr(sys.stdout, "_original_stdstream_copy", normal_stdout) != normal_stdout:
                self.restore_stdout = sys.stdout._original_stdstream_copy
                sys.stdout._original_stdstream_copy = normal_stdout
    
            normal_stderr = sys.__stderr__.fileno()
            self.restore_stderr = None
            if getattr(sys.stderr, "_original_stdstream_copy", normal_stderr) != normal_stderr:
                self.restore_stderr = sys.stderr._original_stdstream_copy
                sys.stderr._original_stdstream_copy = normal_stderr

            self.old_stdout = sys.stdout
            self.old_stderr = sys.stderr
    
            # sys.stdout = open(os.devnull, 'w')
            # sys.stderr = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_value, traceback):
        if self._enabled:
            # sys.stdout = self.old_stdout
            # sys.stderr = self.old_stderr

            if self.restore_stdout is not None:
                sys.stderr._original_stdstream_copy = self.restore_stdout
            if self.restore_stderr is not None:
                sys.stderr._original_stdstream_copy = self.restore_stderr




"""

Model search and download functionality based on llama-cpp-python's:
https://github.com/abetlen/llama-cpp-python/blob/main/docker/open_llama/hug_model.py


Downloads from Hugging Face models hub website:
 https://huggingface.co

 

# example Hugging Face URLS
https://huggingface.co/api/models?author=TheBloke&search=openchat-3.5&tag=
https://huggingface.co/api/models/TheBloke/openchat-3.5-0106-GGUF
https://huggingface.co/{model_id}/resolve/main/{filename}

"""

from typing import Any, Optional, Union, Callable

import requests, json, os, struct, argparse, signal, sys, tempfile
from pprint import pprint

import logging
logger = logging.getLogger(__name__)

from tqdm import tqdm

from sibila import (
    Models,
    GenConf
)
from sibila.models import search_format
from sibila.utils import expand_path



HF_MODEL_SEARCH_URL = "https://huggingface.co/api/models"
HF_MODEL_INFO_URL = "https://huggingface.co/api/models/{model_id}"
HF_MODEL_DOWNLOAD_URL = "https://huggingface.co/{model_id}/resolve/main/{filename}"
HF_ABOUT = "For information about this and other models, please visit https://huggingface.co"


BASE_FORMATS_URL = "https://raw.githubusercontent.com/jndiogo/sibila/main/sibila/base_formats.json"

DOWNLOAD_TIMEOUT = 30
INDENT = " " * 2



def die(message: str,
        exit_code: int = 1):
    print(message,
          file=sys.stderr)
    exit(exit_code)    



def resolve_models_dir(models_dir: Union[str,None],
                       can_default_to_current: bool) -> str:

    if models_dir is None: # check env variable SIBILA_MODELS

        path = os.environ.get(Models.ENV_VAR_NAME)
        if path:
            path_list = path.split(";")
            models_dir = path_list[0]

        elif (os.path.isfile(Models.MODELS_CONF_FILENAME) or
              os.path.isfile(Models.FORMATS_CONF_FILENAME)): # models.json or formats.json in current folder
            models_dir = "./"

        elif can_default_to_current: # safe to default to current folder
            models_dir = "./"

        else:
            die(f"Could not find a 'models' folder in {Models.ENV_VAR_NAME} env variable and can only default to current folder if 'models.json' or 'formats.json' is present.\n"
                f"Please provide the location of a 'models' folder.")


    models_dir = expand_path(models_dir) # type: ignore[arg-type]

    if not os.path.isdir(models_dir):
        die(f"Unable to locate a 'models' directory. Please provide one with -model_dir arg or set env variable {Models.ENV_VAR_NAME}")

    return models_dir




def request_json(url: str, 
                 params: Optional[dict] = None):
    
    logger.info(f"Making request to {url}...")

    response = requests.get(url, params=params)

    if response.status_code == 200:
        return json.loads(response.text)
    else:
        logger.error(f"Request failed with status code {response.status_code}")
        return None
    


def download_file(url: str,
                  destination: str,
                  progress_bar: bool = True,
                  remove_partial: bool = True,
                  remove_on_ctrl_c: bool = True):
    """_summary_

    Args:
        url: _description_
        destination: _description_
        remove_on_error: _description_. Defaults to True.
        remove_on_interruption: _description_. Defaults to True.
    """

    response = requests.get(url,
                            stream=True,
                            timeout=DOWNLOAD_TIMEOUT)

    if response.status_code == 200:
        size_res = response.headers.get("Content-length")
        if size_res:
            size = int(size_res) # type: ignore[assignment]
        else:
            size = -1

        if remove_on_ctrl_c:
            def signal_handler(sig, frame):
                os.remove(destination)
                die(f"Removed partially downloaded file '{destination}'")

            original_handler = signal.getsignal(signal.SIGINT)
            signal.signal(signal.SIGINT, signal_handler)

        try:
            with open(destination, 'wb') as f:

                format = "{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}]"
                with tqdm(total=size,
                          bar_format=format, 
                          disable=not progress_bar,
                          leave=False) as pbar:

                    for chunk in response.iter_content(chunk_size=64 * 1024):
                        if chunk:  # filter out keep-alive new chunks
                            pbar.update(len(chunk))
                            f.write(chunk)

        except Exception as e:
            print(f"Download error:\n{e}")
            if remove_partial:
                os.remove(destination)
                die(f"Removed partially downloaded file '{destination}'")

        if remove_on_ctrl_c:
            signal.signal(signal.SIGINT, original_handler)

    else:
        die(f"Unable to download: status code {response.status_code}")





def search_hub_models(query: str,
                      filter_model_id_gguf: Optional[str] = None,
                      filter_filename: Optional[str] = None,
                      filter_filename_gguf: bool = True,
                      author: Optional[str] = None,
                      tags: Optional[str] = None) -> dict:
    """Search for models in Hugging Face model hub.

    Args:
        query: Text query to search for.
        max_models: Maximum count to return. Use -1 to return {} if more than 1 result is found. Defaults to 10e10.
        author: Optional author filter. Defaults to None.
        tags: Optional tags filter. Defaults to None.

    Returns:
        Dictionary where keys are model names, each entry contains "filenames": [] and "tags": []. Returns {} if not found.
    """
    if filter_filename is not None:
        filter_filename = filter_filename.lower()

    # query parameters
    params = {
        "search": query
    }
    if author is not None:
        params["author"] = author
    if tags is not None:
        params["tags"] = tags

    models = request_json(HF_MODEL_SEARCH_URL, params=params)
    logger.debug(f"HF_MODEL_SEARCH_URL params={params}:\n{models}")

    if models is None:
        return {}
    
    model_dic = {}

    # iterate over the models
    for model in models:
        model_id = model['id']

        if filter_model_id_gguf:
            if "gguf" not in model_id.lower():
                continue

        model_info = request_json(HF_MODEL_INFO_URL.format(model_id=model_id))

        logger.debug(f"HF_MODEL_INFO_URL model_id={model_id}:\n{model_info}")

        # print(model_info['siblings'])
        if model_info is None:
            continue

        entry = {"filenames": [],
                 "tags": model_info.get('tags', [])
                 }

        for sibling in model_info.get('siblings', []):
            filename = sibling.get('rfilename')
            if filename:
                if filter_filename_gguf and not filename.lower().endswith(".gguf"):
                    continue

                if filter_filename and filter_filename not in filename.lower():
                    continue

                entry["filenames"].append(filename)

        if len(entry["filenames"]):
            model_dic[model_id] = entry


    # sort by model name
    model_dic = dict(sorted(model_dic.items()))

    return model_dic




def list_hub_models_and_filenames(models_dic: dict) -> str:
    out = ""
    for model_id,dic in models_dic.items():
        out += f"{model_id}\n"
        for filename in dic["filenames"]:
            out += f"{INDENT}{filename}\n"
        out += "\n"
    return out



def check_res_name(name: str):
    if ":" not in name:
        die(f"Model res_name is not including 'provider:'. Did you mean 'llamacpp:{name}' or 'openai:{name}'?")



# ========================================================================= models

def models(args):

    can_default_to_current = bool(args.test_name)

    models_dir = resolve_models_dir(args.models_dir,
                                    can_default_to_current=can_default_to_current)
    
    print(f"Using models directory '{models_dir}'")

    Models.setup(models_dir,
                 clear=True,
                 add_cwd=False,
                 load_base=True,
                 load_from_env=False)


    if args.set_resname_name_format: # sibila models -s res_name model_name_or_filename [format]

        if len(args.set_resname_name_format) < 2 or len(args.set_resname_name_format) > 3:
            die("Option -s requires 2 or 3 arguments: -s res_name name_or_filename [format]")

        res_name = args.set_resname_name_format[0]
        check_res_name(res_name)
        model_name = args.set_resname_name_format[1]

        if res_name.startswith("llamacpp:"):
            model_name = os.path.basename(model_name)

        format = args.set_resname_name_format[2] if len(args.set_resname_name_format) == 3 else None

        try:
            Models.set_model(res_name=res_name,
                             model_name=model_name,
                             format_name=format)
        
            path = Models.save_models()

            format_disp = "" if format is None else f", format='{format}'"
            print(f"Set model '{res_name}' with name='{model_name}'{format_disp} at '{path}'.")

        except Exception as e:
            die(f"Error: {e}")

    elif args.set_resname_link: # sibila models -sl res_name linked_name

        check_res_name(args.set_resname_link[0])

        try:
            Models.set_model_link(*args.set_resname_link)
            
            path = Models.save_models()
            print(f"Set model '{args.set_resname_link[0]}' linking to '{args.set_resname_link[1]}' at '{path}'.")

        except Exception as e:
            die(f"Error: {e}")


    elif args.format_resname_name: # sibila models -f res_name format

        res_name = args.format_resname_name[0]
        check_res_name(res_name)
        
        format_name = args.format_resname_name[1]
        try:
            Models.update_model(res_name,
                                model_name=None,
                                format_name=format_name,
                                genconf=None)

            path = Models.save_models()
            print(f"Updated model '{args.format_resname_name[0]}' with format '{args.format_resname_name[1]}' at '{path}'.")

        except Exception as e:
            die(f"Error: {e}")


    elif args.delete_name: # sibila models -d res_name 

        check_res_name(args.delete_name)

        try:
            Models.delete_model(args.delete_name)

            path = Models.save_models()
            print(f"Deleted model entry '{args.delete_name}' at '{path}'.")

        except Exception as e:
            die(f"Error: {e}")

    elif args.test_name: # sibila models -t res_name 

        check_res_name(args.test_name)

        try:
            print(f"Testing model '{args.test_name}'...")
            model = Models.create(args.test_name) # noqa: F841
            del model
            
            print(f"Model '{args.test_name}' was properly created and should run fine.")

        except Exception as e:
            die(f"Error: {e}")


    elif args.list_query or ( # horrible but necessary hack follows:
        "-l" in sys.argv or 
        "--list" in sys.argv):

        list_query = args.list_query if args.list_query is not None else ""

        if ":" in list_query:
            prov,name = list_query.split(":")
        else:
            prov = ""
            name = list_query

        dic = Models.list_models(name,
                                 [prov] if prov else [],
                                 False)
        
        if dic:
            last_prov = None
            for res_name,val in dic.items():

                prov,name = res_name.split(":")
                if prov != last_prov:
                    print(f"\n{prov}:")
                    last_prov = prov

                print(INDENT, end="")
                print(f"{name}: ", end="")

                if isinstance(val, dict):
                    print(val)
                else: # link
                    print(f"-> {val}")

            print()
        else:
            print(f"No models found for query '{args.list_query}'")
        
    else:
        assert False, f"Impossible option in {args}"



# ========================================================================= formats

def formats(args):
    models_dir = resolve_models_dir(args.models_dir,
                                    can_default_to_current=False)
    
    print(f"Using models directory '{models_dir}'")

    Models.setup(models_dir,
                 clear=True,
                 add_cwd=False,
                 load_base=True,
                 load_from_env=False)


    if args.query_name: # sibila formats -q query

        res = Models.match_format_entry(args.query_name)

        if res is None:
            print(f"No format found for name/filename '{args.query_name}'")
        else:
            name, dic = res
            print(f"Name/filename '{args.query_name}' matches format entry '{name}'.\nFormat entry '{name}':")
            print(INDENT + f"match: '{dic['match']}'")
            template = dic['template'].replace("\n", "\\n")
            print(INDENT + f"template: █{template}█")


    elif args.set_name_match_template: # sibila formats -s entry_name match template

        try:
            name = args.set_name_match_template[0]
            match = args.set_name_match_template[1]
            template = args.set_name_match_template[2]

            if '{{' not in template: # filename or link
                if os.path.isfile(template):
                    with open(template, "r", encoding="utf-8") as f:
                        template = f.read()
                elif Models.get_format_entry(template) is None:
                    raise ValueError("Can't understand template arg. It can be a Jinja2 text template, its filename or a link to an existing format entry (whose template will be used).")
                
            Models.set_format(name,
                              match,
                              template)
            
            path = Models.save_formats()
            template_disp = template if len(template) <= 24 else template[:24] + "..."
            print(f"Set format '{name}' with match='{match}', template='{template_disp}'")

        except Exception as e:
            die(f"Error: {e}")

    elif args.set_name_link: # sibila formats -sl entry_name linked_name

        try:
            Models.set_format_link(*args.set_name_link)
            
            path = Models.save_formats()
            print(f"Set format '{args.set_name_link[0]}' linking to '{args.set_name_link[1]}' at '{path}'.")

        except Exception as e:
            die(f"Error: {e}")


    elif args.delete_name: # sibila formats -d name 

        try:
            Models.delete_format(args.delete_name)

            path = Models.save_formats()
            print(f"Deleted format entry '{args.delete_name}' at '{path}'.")

        except Exception as e:
            die(f"Error: {e}")


    elif args.update: # sibila formats -u
        with tempfile.TemporaryDirectory() as tmp:

            tmp_path = os.path.join(tmp, "formats.json")
            
            download_file(BASE_FORMATS_URL, 
                          tmp_path,
                          progress_bar=False)
            
            Models.merge_from(tmp_path,
                              preserve_current=True)

        path = Models.save_formats()

        print(f"""\
Updated '{path}' from GitHubs 'sibila/base_formats.json', any existing local entries were preserved.
Total entries: {len(Models.format_dir)}""")


    elif args.list_query or ( # horrible but necessary hack follows:
        "-l" in sys.argv or 
        "--list" in sys.argv): # sibila formats -l [query]

        dic = Models.list_formats(args.list_query if args.list_query is not None else "", 
                                  False)
        if dic:
            for name,val in dic.items():

                print(f"{name}: ", end="")

                if isinstance(val, dict):
                    print(val)
                else: # link
                    print(f"-> {val}")

            print()

        else:
            print(f"No formats found for query '{args.list_query}'")


    else:
        assert False, f"Impossible option in {args}"




# ========================================================================= hub

def hub(args):
    has_set = args.set_resname_format is not None

    if has_set and args.model_download is None:
        die("Option error: -s/--set can only be used with option -d/--download.")


    if args.model_download: # sibila hub -d model_id -f filename -a exact_author -s set name

        models_dir = resolve_models_dir(args.models_dir,
                                        # args.set requires an explicit model_dir, to avoid mistakes
                                        can_default_to_current=not has_set)
        
        print("Searching...")
            
        query = args.model_download
        models_dic = search_hub_models(query, 
                                       author=args.author,
                                       filter_filename=args.filename,
                                       filter_model_id_gguf=True,
                                       filter_filename_gguf=True)
        
        logger.debug(f"search_models():\n{models_dic}")

        if len(models_dic) == 0:
            die(f"Could not find any model matching  query '{query}'.\n" +
                "Run 'sibila hf -l query' to find model names and filenames.")
            
        elif len(models_dic) > 1:
            mlist = list_hub_models_and_filenames(models_dic)

            die(f"Found multiple models matching query '{query}':\n\n{mlist}" +
                "Restrict terms in your query or use -a author and -f filename.\n" + 
                HF_ABOUT)

        model_id = next(iter(models_dic))
        filenames = models_dic[model_id]["filenames"]

        if len(filenames) > 1:
            flist = INDENT + ("\n" + INDENT).join(filenames)

            die(f"Multiple filenames found:\n\n{flist}\n\n" +
                "Filter one of the above with -f 'filename_filter'. For example: -f Q4_K_M\n" + 
                HF_ABOUT)

        filename = filenames[0]

        logger.debug(f"HF_MODEL_DOWNLOAD_URL model_id={model_id} filename={filename}")

        url = HF_MODEL_DOWNLOAD_URL.format(model_id=model_id,
                                           filename=filename)
    
        path = os.path.join(models_dir, filename)

        print(f"Downloading model '{model_id}' file '{filename}' to '{path}'")

        download_file(url, path)

        print(f"\nDownload complete.\n{HF_ABOUT}")

        if args.set_resname_format:

            if len(args.set_resname_format) > 2:
                die("Option -s requires 1 or arguments: -s res_name [format]")
            
            Models.setup(models_dir,
                         clear=True,
                         add_cwd=False,
                         load_base=True,
                         load_from_env=False)

            name = args.set_resname_format[0]
            format_name = args.set_resname_format[1] if len(args.set_resname_format) >= 2 else None

            if not name.startswith("llamacpp:"):
                name = "llamacpp:" + name

            genconf = None
            existing = Models.get_model_entry(name)
            if existing:
                entry = existing[1]
                if format_name is None and "format" in entry:
                    format_name = entry["format"]

                if "genconf" in entry:
                    genconf = GenConf.from_dict(entry["genconf"])


            Models.set_model(res_name=name, 
                             model_name=filename,
                             format_name=format_name,
                             genconf=genconf)


            models_dir_path = os.path.join(models_dir, 'models.json')
            Models.save_models(models_dir_path)

            print(f"\nAlso set model entry with name '{name}' at '{models_dir_path}'.\n")


    elif args.model_list: # sibila hub -l query -f filename -a exact_author 

        print("Searching...")

        models_dic = search_hub_models(args.model_list, 
                                       author=args.author,
                                       filter_model_id_gguf=True,
                                       filter_filename=args.filename,
                                       filter_filename_gguf=True)
        
        if len(models_dic) == 0:
            print("No models found.")
        else:
            text = list_hub_models_and_filenames(models_dic)
            print()
            print(text, end='')
            print(HF_ABOUT)

    else:
        raise ValueError("Unknown option")









def main():


    # create argument parser
    parser = argparse.ArgumentParser(
        prog="sibila",
        description="Sibila CLI tool for managing models and formats.",
        epilog="For help, check https://jndiogo.github.io/sibila/")


    from .__init__ import __version__ as version  # type: ignore[import-not-found]
    parser.add_argument('--version', action='version', version='%(prog)s ' + version)

    subparser = parser.add_subparsers(title="actions",
                                      description="Use 'models' and 'formats' to manage, 'hub' to search and download models.",
                                      help="Run 'sibila {command} --help' for specific help.",
                                      required=True)



    def add_common(parser: Any):
        parser.add_argument("-m", "--models_dir",                             
                            type=str,
                            help="Path to 'models' folder. If not provided, looks for SIBILA_MODELS env variable or defaults to current folder in some (safer) commands.")
        parser.add_argument('-v', '--verbose', 
                            action='count',
                            default=0,
                            help='Verbose output level, Repeat argument for more verboseness')



    # ============================================================================= models
    parser_models = subparser.add_parser('models',
                                         description="Manage a models.json config file.")

    exclusive_models = parser_models.add_mutually_exclusive_group(required=True)
    
    exclusive_models.add_argument('-l', '--list',
                                  metavar="QUERY",
                                  type=str,
                                  nargs="?",
                                  default="",
                                  dest="list_query",
                                  help="List models filtering by model name substring or res_name in the form provider:query. Examples: 'gpt', 'llamacpp:', 'llamacpp:open'.")

    exclusive_models.add_argument('-t', '--test', 
                                  metavar="NAME",
                                  type=str,
                                  dest="test_name",
                                  help="Test if model entry with this res_name can be created. Checks if format template is present.")

    exclusive_models.add_argument('-s', '--set',
                                  metavar="NAME",
                                  type=str,
                                  nargs='+',
                                  dest="set_resname_name_format",
                                  help="Set model entry with 2 args: res_name name_or_filename [format].")
    
    exclusive_models.add_argument('-sl', '--setlink',
                                  metavar="NAME",
                                  type=str,
                                  nargs=2,
                                  dest="set_resname_link",
                                  help="Set model link with 2 args: res_name linked_name.")

    exclusive_models.add_argument('-f', '--format', 
                                  metavar="NAME",
                                  type=str,
                                  nargs=2,
                                  dest="format_resname_name",
                                  help="Update format in existing model, 2 args: res_name format.")

    exclusive_models.add_argument('-d', '--delete', 
                                  metavar="NAME",
                                  type=str,
                                  dest="delete_name",
                                  help="Delete model res_name entry.")


    add_common(parser_models)
    parser_models.set_defaults(func=models)




    # ============================================================================= formats
    parser_formats = subparser.add_parser('formats',
                                          description="Manage a formats.json config file.")

    exclusive_formats = parser_formats.add_mutually_exclusive_group(required=True)
    
    exclusive_formats.add_argument('-l', '--list',
                                   metavar="QUERY",
                                   type=str,
                                   nargs="?",
                                   default="",
                                   dest="list_query",
                                   help="List formats filtering by name substring.")

    exclusive_formats.add_argument('-q', '--query',
                                   metavar="QUERY",
                                   type=str,
                                   dest="query_name",
                                   help="Search for a format from this model name or filename in 'formats.json'.")

    exclusive_formats.add_argument('-s', '--set',
                                   metavar="NAME",
                                   type=str,
                                   nargs=3,
                                   dest="set_name_match_template",
                                   help="Set format entry with 3 args: name, match_regex, template. Arg template can be a Jinja2 template string or filename, or a name of an existing format entry to use as template.")
    
    exclusive_formats.add_argument('-sl', '--setlink',
                                   metavar="NAME",
                                   type=str,
                                   nargs=2,
                                   dest="set_name_link",
                                   help="Set format link with 2 args: name, linked_name.")


    exclusive_formats.add_argument('-d', '--delete', 
                                   metavar="NAME",
                                   type=str,
                                   dest="delete_name",
                                   help="Delete format entry.")


    exclusive_formats.add_argument('-u', '--update', 
                                   action="store_true",
                                   help="Update formats.json from GitHub's 'sibila/base_formats.json', preserving existing local entries.")
    
    add_common(parser_formats)
    parser_formats.set_defaults(func=formats)




    # ============================================================================= hub
    parser_hf = subparser.add_parser('hub',
                                     description="Search for and download models from Hugging Face hub.")

    exclusive_hf = parser_hf.add_mutually_exclusive_group(required=True)

    exclusive_hf.add_argument('-d', '--download', 
                              metavar="MODEL_ID",
                              type=str,
                              dest="model_download",
                              help="Model id or name. Query is case-insensitive, allows multiple words separated with space. For example: TheBloke/openchat-3.5-1210-GGUF")

    exclusive_hf.add_argument('-l', '--list', 
                              metavar="QUERY",
                              type=str,
                              dest="model_list",
                              help="List model names matching query. Query is case-insensitive, allows multiple words separated with space. For example: openchat-3.5. ")


    parser_hf.add_argument("-f", "--filename", 
                           type=str, 
                           help='Case-insensitive filename substring match, for example: Q4_K_M')

    parser_hf.add_argument('-a', '--author', 
                           metavar="EXACT_AUTHOR",
                           type=str, 
                           help="Only from this author. Exact, case-sensitive author name. For example: TheBloke")

    parser_hf.add_argument("-s", "--set", 
                           metavar="NAME",
                           type=str,
                           nargs='+',
                           dest="set_resname_format",
                           help="Set model entry with args: res_name [format]. Use only with option -d/--download.")

    add_common(parser_hf)
   
    parser_hf.set_defaults(func=hub)

    


    args = parser.parse_args()

    if vars(args).get("verbose"):
        logging.basicConfig(level=logging.INFO if args.verbose == 1 else logging.DEBUG,
                            force=True)
        print(vars(args))
    
    args.func(args)
    exit(0)




if __name__ == '__main__':
    main()



from typing import Any, Optional, Union
from enum import IntFlag
from collections.abc import Sequence

import json
from copy import copy
from pprint import pformat

import logging
logger = logging.getLogger(__name__)

from .model import (
    GenConf,
    Model,    
)

from .thread import (
    Thread,
    MsgKind
)



class Trim(IntFlag):
    """Flags for Thread trimming."""

    NONE = 0
    """No trimming."""

    INST = 1
    """Can remove INST message."""

    IN = 2
    """Can remove IN messages."""

    OUT = 4
    """Can remove OUT messages."""
        
    KEEP_FIRST_IN = 1024
    """If trimming IN messages, never remove first one."""

    KEEP_FIRST_OUT = 2048
    """If trimming OUT messages, never remove first one."""



class Context(Thread):

    """A class based on Thread that manages total token length, so that it's kept under a certain value.
    Also supports a persistent inst (instructions) text."""

    
    def __init__(self,                 
                 t: Optional[Union[Thread,list,str,dict,tuple]] = None, 
                 max_token_len: Optional[int] = None,        
                 pinned_inst_text: str = "",
                 join_sep: str = "\n"):
        """
        Args:
            t: Can initialize from a Thread, from a list (containing messages in any format accepted in _parse_msg()) or a single message as an str, an (MsgKind,text) tuple or a dict. Defaults to None.
            max_token_len: Maximum token count to use when trimming. Defaults to None, which will use max model context length.
            pinned_inst_text: Pinned inst text which survives clear(). Defaults to "".
            join_sep: Separator used when message text needs to be joined. Defaults to "\\n".
        """

        super().__init__(t,
                         inst=pinned_inst_text,
                         join_sep=join_sep)
        
        self.max_token_len = max_token_len

        self.pinned_inst_text = pinned_inst_text
        
   
    
    def clear(self):
        """Delete all messages but reset inst to a pinned text if any."""
        super().clear()        
        if self.pinned_inst_text is not None:
            self.inst = self.pinned_inst_text
            


    def trim(self,
             trim_flags: Trim,
             model: Model,
             *,
             max_token_len: Optional[int] = None,
             ) -> bool:
        """Trim context by selectively removing older messages until thread fits max_token_len.

        Args:
            trim_flags: Flags to guide selection of which messages to remove.
            model: Model that will process the thread.
            max_token_len: Cut messages until size is lower than this number. Defaults to None.

        Raises:
            RuntimeError: If unable to trim anything.

        Returns:
            True if any context trimming occurred.
        """

        if max_token_len is None:
            max_token_len = self.max_token_len
            
        if max_token_len is None:
            max_token_len = model.ctx_len

        # if genconf is None:
        #     genconf = model.genconf            
        # assert max_token_len < model.ctx_len, f"max_token_len ({max_token_len}) must be < model's context size ({model.ctx_len}) - genconf.max_new_tokens"
        
        if trim_flags == Trim.NONE: # no trimming
            return False
        
        thread = self.clone()

        any_trim = False
        
        while True:

            curr_len = model.token_len(thread)

            if curr_len <= max_token_len:
                break

            logger.debug(f"len={curr_len} / max={max_token_len}")

            if self.inst and trim_flags & Trim.INST:
                self.inst = ''
                any_trim = True
                logger.debug(f"Cutting INST {self.inst[:80]} (...)")
                continue

            # cut first possible message, starting from oldest first ones
            trimmed = False
            in_index = out_index = 0

            for index,m in enumerate(thread):
                kind,text = m

                if kind == MsgKind.IN:
                    if trim_flags & Trim.IN:
                        if not (trim_flags & Trim.KEEP_FIRST_IN and in_index == 0):
                            del thread[index]
                            trimmed = True
                            logger.debug(f"Cutting IN {text[:80]} (...)")
                            break
                    in_index += 1

                elif kind == MsgKind.OUT:
                    if trim_flags & Trim.OUT:                        
                        if not (trim_flags & Trim.KEEP_FIRST_OUT and out_index == 0):
                            del thread[index]
                            trimmed = True
                            logger.debug(f"Cutting OUT {text[:80]} (...)")
                            break
                    out_index += 1
                
            if not trimmed:
                # all thread messages were cycled but not a single could be cut, so size remains the same
                # arriving here we did all we could for trim_flags but could not remove any more
                raise RuntimeError("Unable to trim anything out of thread")
            else:
                any_trim = True

        # while end
        

        if any_trim:
            self._msgs = thread._msgs
            
        return any_trim






"""Structured data from local or remote LLM models."""

__version__ = "0.3.2"

__all__ = [
    "Models",
    "Model", "TextModel", "MessagesModel", "Tokenizer",
    "GenConf", "GenRes", "GenError", "GenOut",
    "LlamaCppModel", "LlamaCppTokenizer",
    "OpenAIModel", "OpenAITokenizer",
    "Thread", "MsgKind",
    "Context", "Trim",
    "JSchemaConf",
    "TDesc"
]

__author__ = "Jorge Diogo"


from .gen import (
    GenConf,
    GenRes,
    GenError,
    GenOut
)

from .thread import (
    Thread,
    MsgKind,
)

from .model import (
    Model,
    TextModel,
    FormattedTextModel,
    MessagesModel,
    Tokenizer
)

from .llamacpp import (
    LlamaCppModel,
    LlamaCppTokenizer
)

from .openai import (
    OpenAIModel,
    OpenAITokenizer
)

from .context import (
    Context,
    Trim
)

from .models import Models

from .json_schema import JSchemaConf

""" A shorter 'Annotated', meant to be used as: TDesc[type, "description"] """
from typing import Annotated
TDesc = Annotated



import pytest

from typing import Any, Optional, Union, Literal, Annotated, get_origin, get_args
import enum

from datetime import date, time, datetime, timezone

from dataclasses import dataclass, is_dataclass, fields, MISSING

import logging
logging.basicConfig(level=logging.DEBUG)

from pydantic import BaseModel, Field

from sibila.json_schema import (
    JSchemaConf
)

from sibila.null import NullModel


model = NullModel()


def extract(target: Any,
             text: str,
             response: str,
             result: Any,
             *,
             inst: Optional[str]=None):
    model.set_response(response)
    assert model.extract(target, text, inst=inst) == result

def classify(target: Any,
             text: str,
             response: str,
             result: Any,
             *,
             inst: Optional[str]=None):
    model.set_response(response)
    assert model.classify(target, text, inst=inst) == result






def test_extract_root_types():

    extract(bool,
            "It's a great time to surf",
            '{"output": true}',
            True)

    extract(bool,
            "I'll never do it",
            '{"output": false}',
            False)

    extract(list[Literal["dog", "horse", "car", "bus"]], 
             "Dogs and horses and tractors and a school bus",
             '{"output": ["dog", "horse", "car", "bus"]}',
             ['dog', 'horse', 'car', 'bus'])


    from enum import Enum
    class Tag(Enum):
        DOG = "dog"
        HORSE = "horse"
        CAR = "car"
        BUS = "bus"
        OTHER = "other"

    extract(list[Tag],
            "Dogs and horses and a ball",
            '{"output": ["dog", "horse", "other"]}',
            [Tag('dog'), Tag('horse'), Tag('other')],
            inst="Select 'other' if no other category fits")


    extract(list[int],
            "21 years and ten ponies. Five elephants?",
            '{"output": [21, 10, 5]}',
            [21, 10, 5],
            inst="Extract numbers")





    extract(datetime, 
            "Eleven of May, 1984, 10 past 15 AM.", 
            '{"output":"1984-05-11T15:10:00+00:00"}',
            datetime(1984, 5, 11, 15, 10, tzinfo=timezone.utc),
            inst="Date and time in ISO8601 format")




    @dataclass
    class Inv:
        """Class for keeping track of an item in inventory."""
        name: str
        unit_price: float
        alistia: list[int]
        anno: Annotated[str, "This is anno"] = "30"
        quantity_on_hand: int = 0

    extract(Inv,
            "A lot of nice things but no meaning at all",
            '{ "name": "Item", "unit_price": 0, "alistia": [], "anno": "30", "quantity_on_hand": 0}',
            Inv(name='Item', unit_price=0, alistia=[], anno='30', quantity_on_hand=0))





    class UserDetail(BaseModel):
        """ Details about a user """
        name: str
        age: int

    extract(list[UserDetail],
            "Paul and his buddy Mathilda, they love to ride. He's 68 years old, she's 81.",
            '{"output": [{"name": "Paul", "age": 68}, {"name": "Mathilda", "age": 81}]}',
            [UserDetail(name='Paul', age=68), UserDetail(name='Mathilda', age=81)])

    class UserDetail(BaseModel):
        """ Details about a user """
        name: str = Field(..., description="The name for the user")
        age: int = Field(..., description="The user's age")

    extract(list[UserDetail],
            "Jane's 99 years old, Paul is 75.",
            '{"output": [{"name": "Jane", "age": 99}, {"name": "Paul", "age": 75}]}',
            [UserDetail(name='Jane', age=99), UserDetail(name='Paul', age=75)])






def test_classify():

    classify(["yes", "no"], 
             "I will never make it on time",
             '{"output": "no"}',
             "no")

    classify(["dog", "horse", "car", "bus"], 
             "Dogs and horses and tractors and a school bus",
             '{"output": "bus"}',
             "bus")




"""Tools for model interaction, summarization, etc.
 
- interact(): Interact with model as in a chat, using input().
- loop(): Iteratively append inputs and generate model outputs.
- recursive_summarize(): Recursively summarize a (large) text or text file.
"""


from typing import Any, Optional, Union, Callable

import json
from pprint import pformat
from copy import copy

import logging
logger = logging.getLogger(__name__)


from .gen import (
    GenConf,
    GenRes,
    GenOut
)

from .thread import (
    Thread,
    MsgKind
)

from .context import (
    Trim,
    Context
)

from .model import (
    Model
)

from .text_splitter import RecursiveTextSplitter




TRIM_DEFAULT = Trim.IN | Trim.OUT | Trim.KEEP_FIRST_IN


def loop(callback: Callable[[Union[GenOut,None], Context, Model, GenConf], bool],
         model: Model,
         *,
         inst_text: Optional[str] = None,
         in_text: Optional[str] = None,

         trim_flags: Trim = TRIM_DEFAULT,
         ctx: Optional[Context] = None,

         genconf: Optional[GenConf] = None,
         ) -> Context:
    """Iteratively append inputs and generate model outputs.
    
    Callback should call ctx.add_OUT(), ctx.add_IN() and return a bool to continue looping or not.
    
    If last Thread msg is not MsgKind.IN, callback() will be called with out_text=None.

    Args:
        callback: A function(out, ctx, model) that will be iteratively called with model's output.
        model: Model to use for generating.
        inst_text: text for Thread instructions. Defaults to None.
        in_text: Text for Thread's initial MsgKind.IN. Defaults to None.
        trim_flags: Context trimming flags, when Thread is too long. Defaults to TRIM_DEFAULT.
        ctx: Optional input Context. Defaults to None.
        genconf: Model generation configuration. Defaults to None, defaults to model's.
    """
    
    if ctx is None:
        ctx = Context()
    else:
        ctx = ctx

    if inst_text is not None:
        ctx.inst = inst_text
    if in_text is not None:
        ctx.add_IN(in_text)
    
    if genconf is None:
        genconf = model.genconf

    if ctx.max_token_len is not None: # use from ctx
        max_token_len = ctx.max_token_len
    else: # assume max possible for model context and genconf
        max_token_len = model.ctx_len - genconf.max_tokens

    
    while True:

        if len(ctx) and ctx.last_kind == MsgKind.IN:
            # last is an IN message: we can trim and generate
        
            ctx.trim(trim_flags,
                     model,
                     max_token_len=max_token_len
                     )
       
            out = model.gen(ctx, genconf)
        else:
            out = None # first call
        
        res = callback(out, 
                       ctx, 
                       model,
                       genconf)

        if not res:
            break
            

    return ctx
            


def interact(model: Model,
             *,
             ctx: Optional[Context] = None,
             inst_text: Optional[str] = None,
             trim_flags: Trim = TRIM_DEFAULT,
             
             genconf: Optional[GenConf] = None,
             ) -> Context:
    """Interact with model as in a chat, using input().

    Includes a list of commands: type !? to see help.

    Args:
        model: Model to use for generating.
        ctx: Optional input Context. Defaults to None.
        inst_text: text for Thread instructions. Defaults to None.
        trim_flags: Context trimming flags, when Thread is too long. Defaults to TRIM_DEFAULT.
        genconf: Model generation configuration. Defaults to None, defaults to model's.    

    Returns:
        Context after all the interactions.
    """

    def callback(out: Union[GenOut,None], 
                 ctx: Context, 
                 model: Model,
                 genconf: GenConf) -> bool:

        if out is not None:
            if out.res != GenRes.OK_STOP:
                print(f"***Result={GenRes.as_text(out.res)}***")

            if out.text:
                text = out.text
            else:
                text = "***No text out***"
                
            ctx.add_OUT(text)
            print(text)
            print()

        
        def print_thread_info():
            if ctx.max_token_len is not None: # use from ctx
                max_token_len = ctx.max_token_len
            else: # assume max possible for model context and genconf
                max_token_len = model.ctx_len - genconf.max_tokens

            length = model.token_len(ctx, genconf)
            print(f"Thread token len={length}, max len before next gen={max_token_len}")
            

        
        # input loop ===============================================
        MARKER: str = '"""'
        multiline: str = ""

        while True:

            user = input('>').strip()
        
            if multiline:
                if user.endswith(MARKER):
                    user = multiline + "\n" + user[:-3]
                    multiline = ""
                else:
                    multiline += "\n" + user
                    continue
    
            else:
                if not user:
                    return False # terminate loop
                    
                elif user.startswith(MARKER):
                    multiline = user[3:]
                    continue
                    
                elif user.endswith("\\"):
                    user = user[:-1]
                    user = user.replace("\\n", "\n")
                    ctx.add_IN(user)
                    continue
                    
                elif user.startswith("!"): # a command
                    params = user[1:].split("=")
                    cmd = params[0]
                    params = params[1:]
    
                    if cmd == "inst":
                        ctx.clear()
                        if params:
                            text = params[0].replace("\\n", "\n")
                            ctx.inst = text
                            
                    elif cmd == "add" or cmd == "a":
                        if params:
                            try:
                                path = params[0]
                                ctx.addx(path=path)
                                ct = ctx.last_text
                                print(ct[:500])
                            except FileNotFoundError:
                                print(f"Could not load '{path}'")
                        else:
                            print("Path needed")
                                                
                    elif cmd == 'c':
                        print_thread_info()
                        print(ctx)
                        
                    elif cmd == 'cl':
                        if not params:
                            params.append("ctx.json")
                        try:
                            ctx.load(params[0])
                            print(f"Loaded context from {params[0]}")
                        except FileNotFoundError:
                            print(f"Could not load '{params[0]}'")
                        
                    elif cmd == 'cs':
                        if not params:
                            params.append("ctx.json")
                        ctx.save(params[0])
                        print(f"Saved context to {params[0]}")
    
                    elif cmd == 'tl':
                        print_thread_info()
                        
                    elif cmd == 'i':
                        print(f"Model:\n{model.info()}")
                        print(f"GenConf:\n{genconf}\n")
                        
                        print_thread_info()

                    # elif cmd == 'p':
                    #     print(model.text_from_turns(ctx.turns))
                        
                    # elif cmd == 'to':
                    #     token_ids = model.tokens_from_turns(ctx.turns)
                    #     print(f"Prompt tokens={token_ids}")
                                                
    
                    else:
                        print(f"Unknown command '!{cmd}' - known commands:\n"
                              " !inst[=text] - clear messages and add inst (system) message\n"
                              " !add|!a=path - load file and add to last msg\n"
                              " !c - list context msgs\n"
                              " !cl=path - load context (default=ctx.json)\n"
                              " !cs=path - save context (default=ctx.json)\n"
                              " !tl - thread's token length\n"
                              " !i - model and genconf info\n"
                              ' Delimit with """ for multiline begin/end or terminate line with \\ to continue into a new line\n'
                              " Empty line + enter to quit"
                              )
                        # " !p - show formatted prompt (if model supports it)\n"
                        # " !to - prompt's tokens\n"
    
                    print()
                    continue
    
            # we have a user prompt
            user = user.replace("\\n", "\n")
            break

        
        ctx.add_IN(user)
        
        return True # continue looping
            


    # start prompt loop
    ctx = loop(callback,
               model,
               
               ctx=ctx,
               inst_text=inst_text,
               in_text=None, # call callback for first prompt
               trim_flags=trim_flags)

    return ctx







def recursive_summarize(model: Model,
                        text: Optional[str] = None,
                        path: Optional[str] = None,
                        overlap_size: int = 20) -> str:
    """Recursively summarize a (large) text or text file.
     
    Works by:

    1. Breaking text into chunks that fit models context.
    2. Run model to summarize chunks.
    3. Join generated summaries and jump to 1. - do this until text size no longer decreases.

    Args:
        model: Model to use for summarizing.
        text: Initial text.
        path: --Or-- A path to an UTF-8 text file.
        overlap_size: Size in model tokens of the overlapping portions at beginning and end of chunks.

    Returns:
        The summarized text.
    """

    if (text is not None) + (path is not None) != 1:
        raise ValueError("Only one of text or path can be given")

    if path is not None:
        with open(path, "r", encoding="utf-8") as f:
            text = f.read()

    inst_text = """Your task is to do short summaries of text."""
    in_text = "Summarize the following text:\n"
    ctx = Context(pinned_inst_text=inst_text)

    # split initial text
    max_token_len = model.ctx_len - model.genconf.max_tokens - (model.tokenizer.token_len(inst_text + in_text) + 16) 
    logger.debug(f"Max ctx token len {max_token_len}")
    
    token_len_fn = model.tokenizer.token_len_lambda
    logger.debug(f"Initial text token_len {token_len_fn(text)}") # type: ignore[arg-type]
    
    spl = RecursiveTextSplitter(max_token_len, overlap_size, len_fn=token_len_fn)

    round = 0
    while True: # summarization rounds
        logger.debug(f"Round {round} {'='*60}")
        
        in_list = spl.split(text=text)
        in_len = sum([len(t) for t in in_list])

        logger.debug(f"Split in {len(in_list)} parts, total len {in_len} chars")
        
        out_list = []
        for i,t in enumerate(in_list):
    
            logger.debug(f"{round}>{i} {'='*30}")
            
            ctx.clear()
            ctx.add_IN(in_text)
            ctx.add_IN(t)
    
            out = model.gen(ctx)        
            logger.debug(out)
    
            out_list.append(out.text)

        text = "\n".join(out_list)
        
        out_len = len(text) # sum([len(t) for t in out_list])
        if out_len >= in_len:
            break
        elif len(out_list) == 1:
            break
        else:
            round += 1
            
    return text





"""A bag of general utilities."""

import os

def dict_merge(dest: dict, 
               src: dict):
    """
    Recursive merge of dictionary entries into dest dict
    """
    for key in src:
        if key in dest:
            if isinstance(dest[key], dict) and isinstance(src[key], dict):
                dict_merge(dest[key], src[key])
            else: # not both are dicts: overwrite dest's entry with src's
                dest[key] = src[key]
        else: # copy new entry to dest
            dest[key] = src[key]


def is_subclass_of(cls, base_cls):
    """Safe issubclass what also works for instances"""
    return isinstance(cls,type) and issubclass(cls, base_cls)


def synth_desc(flags: int,
               name: str) -> str:
    """Create a description from a key or variable name.
    For example:
        class_label -> "Class label"
    flags:
        0: just copy
        1: replace _ with space and capitalize()
    """
    if flags & 1:
        name = name.replace("_", " ").capitalize()

    return name



def expand_path(path: str) -> str:
    if '~' in path:
        path = os.path.expanduser(path)

    path = os.path.abspath(path)
    path = os.path.normpath(path) # normalize absolute path
    return path



"""Base model classes.

- Model: Base class for an LLM model, containing most common functionality.
- TextModel: Base class for a model with text-based input/output.
- FormattedTextModel: Base class for a model that uses formatted text (chat templates) for input/output.
- MessagesModel: Base class for a model with message-based input/output.
- Tokenizer: Base tokenizer class to encode and decode tokens, measure text length in tokens, track special tokens.
"""

from typing import Any, Optional, Union, Callable, Annotated, Literal, get_origin, get_args
from abc import ABC, abstractmethod
from enum import IntEnum

import json

from copy import copy
from pprint import pformat

import logging
logger = logging.getLogger(__name__)

from dataclasses import dataclass, is_dataclass

from jsonschema import ( # type: ignore[import-untyped]
    validate as json_schema_validate, 
    ValidationError as json_schema_ValidationError,
    SchemaError as json_schema_SchemaError
)

import jinja2
from jinja2.exceptions import TemplateError
from jinja2.sandbox import ImmutableSandboxedEnvironment
version = jinja2.__version__.split('.')
if int(version[0]) < 3:
    raise ImportError

def jinja_raise_exception(message):
    raise TemplateError(message)            

from pydantic import BaseModel


from .gen import (
    GenConf,
    GenRes,
    GenError,
    GenOut
)

from .thread import (
    Thread,
    MsgKind
)

from .json_schema import (
    JSchemaConf,
    json_schema_massage,
    json_schema_from_pydantic,
    pydantic_obj_from_json,
    build_dataclass_object_json_schema,
    build_root_json_schema,
    get_enum_type,
    get_final_type,
    create_final_instance
)

from .utils import is_subclass_of



    



class Tokenizer(ABC):
    """Base tokenizer class to encode and decode tokens, measure text length in tokens, track special tokens."""

    bos_token_id: Union[int, None] # beginning of sentence
    bos_token: Union[str, None]

    eos_token_id: Union[int, None] # end of sentence
    eos_token: Union[str, None]

    pad_token_id: Union[int, None] # padding
    pad_token: Union[str, None]

    unk_token_id: Union[int, None] # unknown token
    unk_token: Union[str, None]

    
    def special_tokens(self) -> list[Any]:
        return [(self.bos_token_id, self.bos_token),
                (self.eos_token_id, self.eos_token),
                (self.pad_token_id, self.pad_token),
                (self.unk_token_id, self.unk_token)]
    
    def special_tokens_map(self) -> dict:
        return {"bos_token": self.bos_token,
                "eos_token": self.eos_token,
                "pad_token": self.pad_token,
                "unk_token": self.unk_token}
    
    @abstractmethod
    def encode(self, 
               text: str) -> list[int]:
        """Encode text into model tokens. Inverse of Decode().

        Args:
            text: Text to be encoded.

        Returns:
            A list of ints with the encoded tokens.
        """
        

    @abstractmethod
    def decode(self,
               token_ids: list[int],
               skip_special: bool = True) -> str:
        """Decode model tokens to text. Inverse of Encode().

        Args:
            token_ids: List of model tokens.
            skip_special: Don't decode special tokens like bos and eos. Defaults to True.

        Returns:
            Decoded text.
        """

    def token_len(self, 
                  text: str) -> int:
        """Returns token length for given text.

        Args:
            text: Text to be measured.

        Returns:
            Token length for given text.
        """

        tokens = self.encode(text)
        return len(tokens)        

    @property
    def token_len_lambda(self) -> Callable[[str], int]:
        return lambda text: self.token_len(text)


    def __str__(self):
        return f"""\
bos={self.bos_token_id}='{self.bos_token}'
eos={self.eos_token_id}='{self.eos_token}'
pad={self.pad_token_id}='{self.pad_token}'
unk={self.unk_token_id}='{self.unk_token}'"""








class Model(ABC):
    """Model is a base class for an LLM model with common functionality.

    LlamaCppModel, OpenAIModel, etc, derive from this class.
    """

    
    is_local_model: bool
    """Is the model running locally?"""
    
    is_message_model: bool
    """Is communication with the model message-based or text-based/token-based?"""

    tokenizer: Tokenizer
    """Tokenizer used to encode text (even for message-based models)."""

    genconf: GenConf
    """Generation configuration: options used during gen()."""

    json_format_instructors: dict
    """If GenConf.json / GenConf.json_schema is used, these strings are appended to first thread msg of either SYS or IN kind. See initialization below."""

    json_in_dumps_kwargs: dict
    """Object to string formatting options for json.dumps() calls. See initialization below."""

    _ctx_len: int
    """Context size: shared input and output context length."""


    
    def __init__(self,
                 is_local_model: bool,
                 genconf: Union[GenConf, None],
                 schemaconf: Union[JSchemaConf, None],
                 tokenizer: Union[Tokenizer, None]):
        """Initializer for base model type, shared by actual model classes like LlamaCpp, OpenAI, etc.

        Args:
            is_local_model: Is the model running locally?
            genconf: Default generation configuration options, used if generation call doesn't supply one.
            schemaconf: Default configuration for JSON schema validation, used if generation call doesn't supply one.
            tokenizer: Tokenizer used to encode text (even for message-based models).
        """
        
        self.is_local_model = is_local_model
        
        self._ctx_len = 0

        self.tokenizer = tokenizer # type: ignore[assignment]

        if genconf is None:
            self.genconf = GenConf()
        else:
            self.genconf = genconf.clone()
        
        if schemaconf is None:
            self.schemaconf = JSchemaConf()
        else:
            self.schemaconf = schemaconf.clone()



        # set either "json" or "json_schema" key values to None to skip.
        self.json_format_instructors = {
            "json": {
                "bypass_if": ["json"], # bypass appending if all lowercase text values are present in thread
                "append_text": "Output JSON.",
                "sep_count": 2
            },
            "json_schema": {
                "bypass_if": ["json", "schema"],
                "append_text": "Output JSON matching the following schema:\n{{json_schema}}",
                "sep_count": 2
            }
        }

        # text going to model: tight, without \u00xx
        self.json_in_dumps_kwargs = {
            "indent": None,
            "ensure_ascii": False
        } 


    
    # ======================================================== internal generation points    
            

    def gen(self,
            thread: Thread,
            genconf: Optional[GenConf] = None,
            ) -> GenOut:
        """Text generation from a Thread, used by the other model generation methods.
        Doesn't raise an exception if an error occurs, always returns GenOut.

        Args:
            thread: The Thread to use as model input.
            genconf: Model generation configuration. Defaults to None, which uses model's default.

        Raises:
            NotImplementedError: If method was not defined by a derived class.

        Returns:
            A GenOut object with result, generated text, etc.
            The output text is in GenOut.text.
        """
        raise NotImplementedError


    


    
    

    def gen_json(self,
                 json_schema: Union[dict,str,None],
                
                 thread: Thread,
                 genconf: Optional[GenConf] = None,

                 massage_schema: bool = True,
                 schemaconf: Optional[JSchemaConf] = None,
                 ) -> GenOut:
        """JSON/JSON-schema constrained generation, returning a Python dict of values, conditioned or not by a JSON schema.
        Doesn't raise an exception if an error occurs, always returns GenOut.

        Args:
            json_schema: A JSON schema describing the dict fields that will be output. None means no schema (free JSON output).
            thread: The Thread to use as model input.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            massage_schema: Simplify schema. Defaults to True.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to Defaults to None, which uses model's default.

        Returns:
            A GenOut object with result, generated text, etc. The output dict is in GenOut.dic.
        """

        if genconf is None:
            genconf = self.genconf

        if genconf.json_schema is not None and json_schema is not None:
            logger.warn("Both arg json_schema and genconf.json_schema are set: using json_schema arg")

        if json_schema is not None:
            if schemaconf is None:
                schemaconf = self.schemaconf

            logger.debug("JSON schema conf:\n" + pformat(schemaconf))

            if massage_schema:
                if not isinstance(json_schema, dict):
                    json_schema = json.loads(json_schema)

                json_schema = json_schema_massage(json_schema, schemaconf) # type: ignore[arg-type]
                logger.debug("Massaged JSON schema:\n" + pformat(json_schema))

        out = self.gen(thread, 
                       genconf(format="json", 
                               json_schema=json_schema))
        
        return out        
        


    
    
    def gen_dataclass(self,
                      cls: Any, # a dataclass
                      thread: Thread,
                      genconf: Optional[GenConf] = None,
                      schemaconf: Optional[JSchemaConf] = None
                      ) -> GenOut:
        """Constrained generation after a dataclass definition.
        An initialized dataclass object is returned in the "value" field of the returned dict.
        Doesn't raise an exception if an error occurs, always returns GenOut containing the created object.

        Args:
            cls: A dataclass definition.
            thread: The Thread object to use as model input.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Returns:
            A GenOut object with result, generated text, etc. The initialized dataclass object is in GenOut.value.
        """

        if is_dataclass(cls):
            schema = build_dataclass_object_json_schema(cls)
        else:
            raise TypeError("Only dataclass allowed for argument cls")

        out = self.gen_json(schema,
                            thread,
                            genconf,
                            massage_schema=True,
                            schemaconf=schemaconf)
    
        if out.dic is not None:
            try:
                obj = create_final_instance(cls, 
                                            is_list=False,
                                            val=out.dic,
                                            schemaconf=schemaconf)
                out.value = obj
                
            except TypeError as e:
                out.res = GenRes.ERROR_JSON_SCHEMA_VAL # error initializing object from JSON
                out.text += f"\nJSON Schema error: {e}"
        else:
            # out.res already holds the right error
            ...
        
        return out



    def gen_pydantic(self,
                     cls: Any, # a Pydantic BaseModel class
                     thread: Thread,
                     genconf: Optional[GenConf] = None,
                     schemaconf: Optional[JSchemaConf] = None
                     ) -> GenOut:
        """Constrained generation after a Pydantic BaseModel-derived class definition.
        An initialized Pydantic BaseModel object is returned in the "value" field of the returned dict.
        Doesn't raise an exception if an error occurs, always returns GenOut containing the created object.

        Args:
            cls: A class derived from a Pydantic BaseModel class.
            thread: The Thread to use as model input.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Returns:
            A GenOut object with result, generated text, etc. The initialized Pydantic BaseModel-derived object is in GenOut.value.
        """

        if is_subclass_of(cls, BaseModel):
            schema = json_schema_from_pydantic(cls)
        else:
            raise TypeError("Only pydantic BaseModel allowed for argument cls")

        out = self.gen_json(schema,
                            thread,
                            genconf,
                            massage_schema=True,
                            schemaconf=schemaconf)
    
        if out.dic is not None:
            try:
                obj = pydantic_obj_from_json(cls, 
                                             out.dic,
                                             schemaconf=schemaconf)
                out.value = obj
                
            except TypeError as e:
                out.res = GenRes.ERROR_JSON_SCHEMA_VAL # error validating for object (by Pydantic), but JSON is valid for its schema
                out.text += f"\nJSON Schema error: {e}"
        else:
            # out.res already holds the right error
            ...
        
        return out





    def gen_extract(self,
                    target: Any,
                    thread: Thread,
                    genconf: Optional[GenConf] = None,
                    schemaconf: Optional[JSchemaConf] = None
                    ) -> GenOut:
        """Free type constrained generation: an instance of the given type is initialized with the model's output.
        The initialized value is placed in the "value" field of the returned dict.
        Doesn't raise an exception if an error occurs, always returns GenOut.

        The following target types are accepted:

            - prim_type:
                bool
                int
                float
                str
                
            - enums:
                [1, 2, 3] or ["a","b"] - all items of the same prim_type
                Literal['year', 'name'] - all items of the same prim_type
                Enum, EnumInt, EnumStr, (Enum, int),... - all items of the same prim_type

            - datetime/date/time

            - a list in the form:
                list[type] - for example list[int]. 
                The list can be annotated:
                    Annotated[list[T], "List desc"]
                And/or the list item type can be annotated:
                    list[Annotated[T, "Item desc"]]

            - dataclass with fields of the above supported types (or dataclass).

            - Pydantic BaseModel

        All types can be Annotated[T, "Desc"], for example: 
            count: int
        Can be annotated as:
            count: Annotated[int, "How many units?"]

        Args:
            target: One of the above types.
            thread: The Thread to use as model input.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Returns:
            A GenOut object with model's results and instantiated type in the "value" field.
        """

        OUTPUT_KEY_NAME = "output"

        schema, created_output_key = build_root_json_schema(target, OUTPUT_KEY_NAME)
        
        final_type, is_list = get_final_type(target)

        if schemaconf is None:
            schemaconf = JSchemaConf()

        out = self.gen_json(schema,
                            thread,
                            genconf,
                            massage_schema=True,
                            schemaconf=schemaconf)
    
        if out.dic is not None:

            if created_output_key:
                val = out.dic[OUTPUT_KEY_NAME]
            else:
                val = out.dic

            try:
                value = create_final_instance(final_type, 
                                              is_list, 
                                              val,
                                              schemaconf=schemaconf)
                out.value = value

            except TypeError as e:
                out.res = GenRes.ERROR_JSON_SCHEMA_VAL # error validating, but JSON is valid for its schema
                out.text += f"\nJSON Schema error: {e}"

        else:
            # out.res already holds the right error
            ...

        return out





    # ======================================================== user generation points    


    def __call__(self,             
                 query: Union[str,Thread],
                 *,
                 inst: Optional[str] = None,

                 genconf: Optional[GenConf] = None,
                 ok_length_is_error: bool = False
                 ) -> str:
        """Text generation from a Thread or plain text, used by the other model generation methods.

        Args:
            query: Thread or an str with the text of a single IN message to use as model input.
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            ok_length_is_error: Should a result of GenRes.OK_LENGTH be considered an error and raise?

        Raises:
            GenError: If an error occurred. This can be a model error, or an invalid JSON output error.

        Returns:
            Text generated by model.
        """
        
        thread = Thread.ensure(query, inst)

        out = self.gen(thread=thread, 
                       genconf=genconf)
        
        GenError.raise_if_error(out,
                                ok_length_is_error=ok_length_is_error)

        return out.text




    def json(self,             
             json_schema: Union[dict,str,None],
             
             query: Union[str,Thread],
             *,
             inst: Optional[str] = None,

             genconf: Optional[GenConf] = None,
             massage_schema: bool = True,
             schemaconf: Optional[JSchemaConf] = None,
             ) -> dict:
        """JSON/JSON-schema constrained generation, returning a Python dict of values, constrained or not by a JSON schema.
        Raises GenError if unable to get a valid/schema-validated JSON.

        Args:
            json_schema: A JSON schema describing the dict fields that will be output. None means no schema (free JSON output).
            query: Thread or an str with the text of a single IN message to use as model input.
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            massage_schema: Simplify schema. Defaults to True.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            GenError: If an error occurred, for example an invalid JSON schema output error. See GenError.

        Returns:
            A dict from model's JSON response, following genconf.jsonschema, if provided.
        """        

        thread = Thread.ensure(query, inst)

        out = self.gen_json(json_schema,                            
                            thread,
                            genconf,
                            massage_schema,
                            schemaconf)
        
        GenError.raise_if_error(out,
                                ok_length_is_error=False) # as valid JSON can still be produced

        return out.dic # type: ignore[return-value]




    def dataclass(self, # noqa: E811
                  cls: Any, # a dataclass definition

                  query: Union[str,Thread],
                  *,
                  inst: Optional[str] = None,

                  genconf: Optional[GenConf] = None,
                  schemaconf: Optional[JSchemaConf] = None
                  ) -> Any: # a dataclass object
        """Constrained generation after a dataclass definition, resulting in an object initialized with the model's response.
        Raises GenError if unable to get a valid response that follows the dataclass definition.

        Args:
            cls: A dataclass definition.
            query: Thread or an str with the text of a single IN message to use as model input.
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            GenError: If an error occurred, for example invalid object initialization. See GenError.

        Returns:
            An object of class cls (derived from dataclass) initialized from the constrained JSON output.
        """

        thread = Thread.ensure(query, inst)

        out = self.gen_dataclass(cls,
                                 thread,
                                 genconf,
                                 schemaconf)

        GenError.raise_if_error(out,
                                ok_length_is_error=False) # as valid JSON can still be produced

        return out.value






    def pydantic(self,
                 cls: Any, # a Pydantic BaseModel class

                 query: Union[str,Thread],
                 *,
                 inst: Optional[str] = None,

                 genconf: Optional[GenConf] = None,
                 schemaconf: Optional[JSchemaConf] = None
                 ) -> Any: # a Pydantic BaseModel object
        """Constrained generation after a Pydantic BaseModel-derived class definition.
        Results in an object initialized with the model response.
        Raises GenError if unable to get a valid dict that follows the BaseModel class definition.

        Args:
            cls: A class derived from a Pydantic BaseModel class.
            query: Thread or an str with the text of a single IN message to use as model input.
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            GenError: If an error occurred, for example an invalid BaseModel object. See GenError.

        Returns:
            A Pydantic object of class cls (derived from BaseModel) initialized from the constrained JSON output.
        """

        thread = Thread.ensure(query, inst)

        out = self.gen_pydantic(cls,
                                thread,
                                genconf,
                                schemaconf)

        GenError.raise_if_error(out,
                                ok_length_is_error=False) # as valid JSON can still be produced

        return out.value







    def extract(self,
                target: Any,

                query: Union[str,Thread],
                *,
                inst: Optional[str] = None,

                genconf: Optional[GenConf] = None,
                schemaconf: Optional[JSchemaConf] = None
                ) -> Any:
        
        """Free type constrained generation: an instance of the given type will be initialized with the model's output.
        The following target types are accepted:

        - prim_type:

            - bool
            - int
            - float
            - str
            
        - enums:

            - [1, 2, 3] or ["a","b"] - all items of the same prim_type
            - Literal['year', 'name'] - all items of the same prim_type
            - Enum, EnumInt, EnumStr, (Enum, int),... - all items of the same prim_type

        - datetime/date/time

        - a list in the form:
            - list[type]
            
            For example list[int]. The list can be annotated:
                Annotated[list[T], "List desc"]
            And/or the list item type can be annotated:
                list[Annotated[T, "Item desc"]]

        - dataclass with fields of the above supported types (or dataclass).

        - Pydantic BaseModel

        All types can be Annotated[T, "Desc"], for example: 
            count: int
        Can be annotated as:
            count: Annotated[int, "How many units?"]

        Args:
            target: One of the above types.
            query: Thread or an str with the text of a single IN message to use as model input.
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            GenError: If an error occurred, for example invalid object initialization. See GenError.

        Returns:
            A value of target arg type instantiated with the model's output.
        """

        thread = Thread.ensure(query, inst)

        out = self.gen_extract(target,
                               thread,
                               genconf,
                               schemaconf)

        GenError.raise_if_error(out,
                                ok_length_is_error=False) # as valid JSON can still be produced

        return out.value




    def classify(self,
                 labels: Any,

                 query: Union[str,Thread],
                 *,
                 inst: Optional[str] = None,

                 genconf: Optional[GenConf] = None,
                 schemaconf: Optional[JSchemaConf] = None
                 ) -> Any:
        """Returns a classification from one of the given enumeration values
        The following ways to specify the valid labels are accepted:

        - [1, 2, 3] or ["a","b"] - all items of the same prim_type
        - Literal['year', 'name'] - all items of the same prim_type
        - Enum, EnumInt, EnumStr, (Enum, int),... - all items of the same prim_type

        Args:
            labels: One of the above types.
            query: Thread or an str with the text of a single IN message to use as model input.
            inst: Instruction message for model. Will override Thread's inst, if set. Defaults to None.
            genconf: Model generation configuration. Defaults to None, which uses model's default.
            schemaconf: JSchemaConf object that controls schema simplification. Defaults to None, which uses model's default.

        Raises:
            GenError: If an error occurred. See GenError.

        Returns:
            One of the given labels, as classified by the model.
        """

        # verify it's a valid enum "type"
        type_,_ = get_enum_type(labels)
        if type_ is None:
            raise TypeError("Arg labels must be one of Literal, Enum class or a list of str, float or int items")
        
        return self.extract(labels,
                            query,
                            inst=inst,
                            genconf=genconf,
                            schemaconf=schemaconf)











    
    # ======================================================== properties

     

    @abstractmethod
    def token_len(self,
                  thread: Thread) -> int:
        """Calculate token length for a Thread.

        Args:
            thread: For token length calculation.

        Returns:
            Number of tokens the thread will use.
        """
        ...

    
    @property
    def ctx_len(self) -> int:
        """Maximum context length, shared for input + output.
        We assume a common in+out context where total token length must always be less than this number."""
        return self._ctx_len

    @classmethod
    def known_models(cls) -> Union[list[str], None]:
        """If the model can only use a fixed set of models, return their names. Otherwise, return None.

        Returns:
            Returns a list of known models or None if it can accept any model.
        """
        return None

    @classmethod
    def version(cls) -> str:
        """Sibila version + provider version
        Ex: sibila='0.2.3' provider='llama-cpp-python 0.2.44'        
        """        
        from .__init__ import __version__ as version  # type: ignore[import-not-found]
        return f"sibila='{version}' provider='{cls.provider_version()}'"
        
    @classmethod
    @abstractmethod
    def provider_version(cls) -> str:
        """Provider library version: provider x.y.z
        Ex. llama-cpp-python 0.2.44
        """
        ...


    @property
    def desc(self) -> str:
        """Model description."""
        return "Unknown model desc"
    
    def info(self) -> str:
        """Model object information."""
        return f"desc='{self.desc}'," \
               f"ctx_len={self.ctx_len},\n" \
               f"genconf={self.genconf}"
               
    
    def __str__(self):
        return self.info()





    
    
    # = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = internal
    
    def _prepare_gen_in(self,
                        thread: Thread,
                        genconf: GenConf) -> Thread:
        """Perform common pre-generation operations of the thread that will be used.

        Args:
            thread: Input thread that may be augmented here. If not modified, the same thread is returned.
            genconf: Generation configuration.

        Raises:
            ValueError: If empty thread.

        Returns:
            A thread with A GenOut object with result, generated text, etc.
        """

        logger.debug(f"GenConf: {genconf}")

        if not len(thread):
            raise ValueError("Cannot generate from an empty thread")
        
        if thread.last_kind != MsgKind.IN:
            logger.warn("Last thread message is not an IN message.")

        if genconf.format == "json":
            if genconf.json_schema is None:
                inst = self.json_format_instructors["json"]
            else:
                inst = self.json_format_instructors["json_schema"]

            if inst is not None:
                bypass = all([thread.has_text_lower(t) for t in inst["bypass_if"]])
    
                if not bypass:
                    thread = thread.clone()
                    text = inst["append_text"]
                    
                    if genconf.json_schema is not None:
                        if isinstance(genconf.json_schema, str): 
                            sc = genconf.json_schema
                        else:
                            sc = json.dumps(genconf.json_schema, **self.json_in_dumps_kwargs)

                        text = text.replace("{{json_schema}}", sc)

                    first_text = thread.get_text(0)
                    first_text = thread.join_text(first_text, text, inst["sep_count"]) # will separate by join_sep * sep_count
                    thread.set_text(0, first_text)
                    
                    logger.debug(f'Appended {"json" if genconf.json_schema is None else "json_schema"} instruction to first message.')

            
        logger.debug(f"Prepare: {str(thread)}")
        
        return thread


    
    
    def _prepare_gen_out(self,
                         text: str, 
                         finish: str,
                         genconf: GenConf) -> GenOut:
        """Perform common operations over the generated model response.

        Args:
            text: Text obtained from model.
            finish: Model finish reason - only "stop", "length" are used, others will map as GenRes.ERROR_MODEL.

        Returns:
            A GenOut object with result, generated text, etc. 
        """

        logger.debug(f"Response {finish}: █{text}█")
        
        if text is None:
            text = ''
        else:
            text = text.strip()

        if genconf.format == "json":
            if "\\u" in text:
                # dumps(with default ensure_ascii=False) -> ascii (subset of utf-8) -> text.encode("latin1") -> latin1 ->
                #   decode("unicode-escape") -> utf-8
                text = text.encode("latin1").decode("unicode-escape")
                
            try:
                dic = json.loads(text)

                if genconf.json_schema is not None:
                    if isinstance(genconf.json_schema, str):
                        schema_dic = json.loads(genconf.json_schema)
                    else:
                        schema_dic = genconf.json_schema
                        
                    json_schema_validate(dic, schema_dic)

                out = GenOut(res=GenRes.from_finish_reason(finish),
                             text=text,
                             dic=dic)
            
            except json.JSONDecodeError:
                out = GenOut(res=GenRes.from_finish_reason("!json"),
                             text=text)
                
            except json_schema_ValidationError as err:
                logger.info(f"JSON schema validation error: {err.message}")
                out = GenOut(res=GenRes.from_finish_reason("!json_schema_val"),
                             text=text,
                             dic=dic)
                
            except json_schema_SchemaError as err:
                logger.info(f"JSON schema error: {err.message}")
                out = GenOut(res=GenRes.from_finish_reason("!json_schema_error"),
                             text=text,
                             dic=dic)

        else:
            out = GenOut(res=GenRes.from_finish_reason(finish),
                         text=text)
            
       
        return out
    










class TextModel(Model, ABC):
    """Model with text-based input/output."""

    def __init__(self,
                 is_local_model: bool,
                 genconf: Union[GenConf, None],
                 schemaconf: Union[JSchemaConf, None],
                 tokenizer: Union[Tokenizer, None]):

        super().__init__(is_local_model,
                         genconf,
                         schemaconf,
                         tokenizer)
        
        self.is_message_model = False

    
    def token_len(self,
                  thread: Thread,
                  _: Optional[GenConf] = None) -> int:
        """Calculate token length for a Thread.

        Args:
            thread: For token length calculation.

        Returns:
            Number of tokens the thread will use.
        """

        text = self.text_from_thread(thread)
        return self.tokenizer.token_len(text)

   
    
    @abstractmethod
    def text_from_thread(self,
                         thread: Thread) -> str:
        ...
    







from .models import Models


class FormattedTextModel(TextModel, ABC):
    """Model that uses formatted text (chat templates) for input/output."""
    
    format_template: Union[str,None]
    _jinja_compiled_template: Union[Any,None]

    def __init__(self,
                 is_local_model: bool,
                 genconf: Union[GenConf, None],
                 schemaconf: Union[JSchemaConf, None],
                 tokenizer: Union[Tokenizer, None]):

        super().__init__(is_local_model,
                         genconf,
                         schemaconf,
                         tokenizer)

        self.format_template = None
        self._jinja_compiled_template = None
        
    
    def init_format(self,
                    format: Union[str,None],
                    format_search_order: list,
                    info: dict):
        """
        format_search_order is a list of str flags:
            "name: match by name
            "meta_template": use model file metadata's template
            "formats_json": look for a "formats.json" file in the same folder
            ex: ["name", "meta_template", "formats_json"]
        
        info fields: {
            "name": "filename.gguf"
            "path": full path to model
            "meta_template_name": "chat_template"
        }
        """

        # handle chat template format
        def search_format() -> str:

            for order in format_search_order:

                if order == "name":
                    fmt = Models.match_format_template(info["name"])
                    if fmt is not None:
                        return fmt
                        
                elif order == "meta_template":
                    md = self.get_metadata()
                    key_name = info["meta_template_name"]
                    if key_name in md:
                        logger.debug(f"Format from model file metadata ('{key_name}') template='{md[key_name]}'")
                        fmt = md[key_name]
                        return fmt # type: ignore[return-value]
                    
                elif order == "formats_json":
                    fmt = Models.folder_match_format_template(info["path"])
                    if fmt is not None:
                        return fmt

            
            raise ValueError("Could not find a suitable chat template format for this model. "
                             "Without a format, fine-tuned models cannot function properly. "
                             "See the docs on how you can fix this: "
                             "either setup the format in Models factory, or provide the chat template in the 'format' arg.")

        
        if format is not None:
            if '{{' in format: # an str with a jinja template
                self.format_template = format
                
            else: # a format name
                self.format_template = Models.get_format_template(format)

                
        if self.format_template is None:
            self.format_template = search_format() # will raise if unable to find

        # setup jinja template
        jinja_env = ImmutableSandboxedEnvironment(trim_blocks=True, lstrip_blocks=True)
        jinja_env.globals["raise_exception"] = jinja_raise_exception
        self._jinja_compiled_template = jinja_env.from_string(self.format_template)



    
    def gen(self, 
            thread: Thread,
            genconf: Optional[GenConf] = None,
            ) -> GenOut:
        """Text generation from a Thread, used by the other model generation methods.
        Doesn't raise an exception if an error occurs, always returns GenOut.

        Args:
            thread: The Thread object to use as model input.
            genconf: Model generation configuration. Defaults to None, which uses model's default.

        Returns:
            A GenOut object with result, generated text, etc. 
        """

        if genconf is None:
            genconf = self.genconf

        thread = self._prepare_gen_in(thread, genconf)

        prompt = self.text_from_thread(thread)
        
        logger.debug(f"Prompt: █{prompt}█")

        text,finish = self._gen_text(prompt, genconf)

        out = self._prepare_gen_out(text, finish, genconf)

        return out


    
    
    @abstractmethod
    def _gen_text(self,
                  text: str,
                  genconf: GenConf) -> tuple[str,str]:
        """Generate from formatted text.

        Args:
            text: Formatted text (from input Thread).
            genconf: Model generation configuration.

        Returns:
            Tuple of strings: generated_text, finish_reason.
        """
        ...


    
    def text_from_thread(self,
                         thread: Thread) -> str:

        messages = thread.as_chatml()
        text = self._jinja_compiled_template.render(messages=messages, # type: ignore[union-attr]
                                                    add_generation_prompt=True,
                                                    **self.tokenizer.special_tokens_map())
        return text
        

    
    
    def get_metadata(self) -> dict:
        """Returns model metadata."""
        return {}








class MessagesModel(Model, ABC):
    """Model with message-based communication."""

    def __init__(self,
                 is_local_model: bool,
                 genconf: Union[GenConf,None],
                 schemaconf: Union[JSchemaConf,None],
                 tokenizer: Union[Tokenizer, None]):

        super().__init__(is_local_model,
                         genconf,
                         schemaconf,
                         tokenizer)
        
        self.is_message_model = True
                



"""Null model(s) for testing.

Created later if  needed: NullFormattedTextModel, NullMessagesModel
"""


from typing import Any, Optional, Union

import sys, os, json, ctypes
from copy import copy

import logging
logger = logging.getLogger(__name__)


from .gen import (
    GenConf,
    GenOut
)

from .thread import (
    Thread
)

from .model import (
    Model,
    MessagesModel,
    FormattedTextModel,
    Tokenizer
)

from .json_schema import JSchemaConf




class NullModel(Model):

    def __init__(self,
                 is_local_model: bool = True,
                 is_message_model: bool = False,

                 genconf: Optional[GenConf] = None,
                 schemaconf: Optional[JSchemaConf] = None
                 ):

        super().__init__(is_local_model,
                         genconf,
                         schemaconf,
                         None)
        
        self.is_message_model = is_message_model

        self.response = {
            "text": "",
            "finish": "stop"
        }


    def set_response(self,
                     text: str,
                     finish="stop"):
        """
        finish: stop, length
        """
        self.response["text"] = text
        self.response["finish"] = finish



    def gen(self, 
            thread: Thread,
            genconf: Optional[GenConf] = None,
            ) -> GenOut:
        
        if genconf is None:
            genconf = self.genconf

        out = self._prepare_gen_out(self.response["text"],
                                    self.response["finish"], 
                                    genconf)

        return out



    def token_len(self,
                  thread: Thread,
                  _: Optional[GenConf] = None) -> int:
        assert False, "NullModel doesn't use tokens"


    @classmethod
    def provider_version(_) -> str:
        """Provider library version: provider x.y.z
        Ex. openai 1.3.6
        """
        return "NullModel 0.0.0"



"""JSON Schema grammar generation utilities for local models."""

from typing import Any, Optional, Union, Literal, Annotated, get_origin, get_args

import json, re

import logging
logger = logging.getLogger(__name__)



# ============================================================================ Grammars

"""
llama.cpp json-schema to grammar converter adapted from llama_cpp_python:
https://github.com/abetlen/llama-cpp-python

Originally from llama.cpp:
https://github.com/ggerganov/llama.cpp/blob/master/examples/json-schema-to-grammar.py
https://github.com/ggerganov/llama.cpp/tree/master/grammars
"""


# whitespace is constrained to a single space char to prevent model "running away" in
# whitespace. Also maybe improves generation quality?
SPACE_RULE = '" "?'

PRIMITIVE_RULES = {
    "boolean": '("true" | "false") space',
    "number": '("-"? ([0-9] | [1-9] [0-9]*)) ("." [0-9]+)? ([eE] [-+]? [0-9]+)? space',
    "integer": '("-"? ([0-9] | [1-9] [0-9]*)) space',
    "string": r""" "\"" (
        [^"\\\x7F\x00-\x1F] |
        "\\" (["\\/bfnrt] | "u" [0-9a-fA-F] [0-9a-fA-F] [0-9a-fA-F] [0-9a-fA-F])
      )* "\"" space """,
    "null": '"null" space',
}

INVALID_RULE_CHARS_RE = re.compile(r"[^a-zA-Z0-9-]+")
GRAMMAR_LITERAL_ESCAPE_RE = re.compile(r'[\r\n"]')
GRAMMAR_LITERAL_ESCAPES = {"\r": "\\r", "\n": "\\n", '"': '\\"'}


class SchemaConverter:
    _prop_order: dict
    _rules: dict
    _defs: dict

    def __init__(self, 
                 prop_order: dict):
        self._prop_order = prop_order
        self._rules = {"space": SPACE_RULE}
        self._defs: dict[str, Any] = {}

    def _format_literal(self, 
                        literal: str):
        escaped: str = GRAMMAR_LITERAL_ESCAPE_RE.sub(
            lambda m: GRAMMAR_LITERAL_ESCAPES.get(m.group(0)), json.dumps(literal) # type: ignore[arg-type,return-value]
        )
        return f'"{escaped}"'

    def _add_rule(self, 
                  name: str, 
                  rule: str):
        esc_name = INVALID_RULE_CHARS_RE.sub("-", name)
        if esc_name not in self._rules or self._rules[esc_name] == rule:
            key = esc_name
        else:
            i = 0
            while f"{esc_name}{i}" in self._rules:
                i += 1
            key = f"{esc_name}{i}"
        self._rules[key] = rule
        return key

    def visit(self, 
              schema: dict[str, Any], 
              name: str) -> str:
        
        rule_name = name or "root"

        if "$defs" in schema:
            # add defs to self._defs for later inlining
            for def_name, def_schema in schema["$defs"].items():
                self._defs[def_name] = def_schema

        if "oneOf" in schema or "anyOf" in schema:
            rule = " | ".join(
                (
                    self.visit(alt_schema, f'{name}{"-" if name else ""}{i}')
                    for i, alt_schema in enumerate(
                        schema.get("oneOf") or schema["anyOf"]
                    )
                )
            )
            return self._add_rule(rule_name, rule)

        elif "const" in schema:
            return self._add_rule(rule_name, self._format_literal(schema["const"]))

        elif "enum" in schema:
            rule = " | ".join((self._format_literal(v) for v in schema["enum"]))
            return self._add_rule(rule_name, rule)

        elif "$ref" in schema:
            ref = schema["$ref"]
            assert ref.startswith("#/$defs/"), f"Unrecognized schema: {schema}"
            # inline $defs
            def_name = ref[len("#/$defs/"):]
            def_schema = self._defs[def_name]
            return self.visit(def_schema, f'{name}{"-" if name else ""}{def_name}')


        schema_type: Optional[str] = schema.get("type") # type: ignore
        assert isinstance(schema_type, str), f"Unrecognized schema: {schema}"

        if schema_type == "object" and "properties" in schema:
            
            if len(self._prop_order):
                prop_pairs = dict(sorted(
                    schema["properties"].items(),
                    # sort by position in prop_order (if specified) then by key
                    key=lambda kv: (self._prop_order.get(kv[0], len(self._prop_order)), kv[0]),
                ))
            else:
                prop_pairs = schema["properties"]

            
            # split names into required, not_required
            required: list = schema.get("required") or []
            not_required = [n for n in prop_pairs if n not in required]
            
            if len(required) == 0: # force all to be required: or the leading comma for not_required may cause broken JSON
                logger.debug(f"Rule '{rule_name}': GBNF grammar cannot parse rule with only optional items: making all items required")
                required = not_required
                not_required = []

            
            def emit_prop(prop_name: str, 
                          prop_schema: dict[str, Any],
                          is_required: bool,
                          index: int
                          ) -> str:
                
                prop_rule_name = self.visit(
                    prop_schema,
                    f'{name}{"-" if name else ""}{prop_name}'
                )

                out = ''                
                
                if not is_required:
                    out += ' ('
                    
                if index > 0:
                    out += ' "," space'

                out += rf' {self._format_literal(prop_name)} space ":" space {prop_rule_name}'

                if not is_required:
                    out += ' )?'

                return out
                

            rule = '"{" space'

            index = 0 # used to add separating "," - only 

            # requireds first
            for prop_name in required:
                prop_schema = prop_pairs[prop_name]
                
                rule += emit_prop(prop_name, prop_schema, 
                                  True, index)
                index += 1

            # non-requireds then
            for prop_name in not_required:
                prop_schema = prop_pairs[prop_name]
                
                rule += emit_prop(prop_name, prop_schema, 
                                  False, index)
                index += 1            
            
            rule += ' "}" space'

            return self._add_rule(rule_name, rule)
            

        elif schema_type == "array" and "items" in schema:
            # TODO `prefixItems` keyword
            item_rule_name = self.visit(
                schema["items"], f'{name}{"-" if name else ""}item'
            )
            rule = (
                f'"[" space ({item_rule_name} ("," space {item_rule_name})*)? "]" space'
            )
            return self._add_rule(rule_name, rule)

        else:
            assert schema_type in PRIMITIVE_RULES, f"Unrecognized schema: {schema}"
            return self._add_rule(
                "root" if rule_name == "root" else schema_type,
                PRIMITIVE_RULES[schema_type],
            )

    def format_grammar(self):
        return "\n".join((f"{name} ::= {rule}" for name, rule in self._rules.items()))


def gbnf_from_json_schema(schema: Union[str,dict],
                          prop_order: list[str] = []):
    
    """
    prop_order sorting is probably a bad idea, because it makes output order different from the schema example order, which may unnecessarily confuse the model
    """
    out_schema: dict[str, Any]
    if isinstance(schema, str):
        out_schema = json.loads(schema)
    else:
        out_schema = schema
        
    prop_order_inv: dict[str,int] = {name: idx for idx, name in enumerate(prop_order)}
    
    converter = SchemaConverter(prop_order_inv)
    converter.visit(out_schema, "")
    
    return converter.format_grammar()



"""
JSON_GBNF from llama.cpp:
https://github.com/ggerganov/llama.cpp/tree/master/grammars
string rule altered to disallow raw \n inside ""
"""

JSON_GBNF = r"""
root   ::= object
value  ::= object | array | string | number | ("true" | "false" | "null") ws

object ::=
  "{" ws (
            string ":" ws value
    ("," ws string ":" ws value)*
  )? "}" ws

array  ::=
  "[" ws (
            value
    ("," ws value)*
  )? "]" ws

string ::=
  "\"" (
    [^"\\\x7F\x00-\x1F] |
    "\\" (["\\/bfnrt] | "u" [0-9a-fA-F] [0-9a-fA-F] [0-9a-fA-F] [0-9a-fA-F]) # escapes
  )* "\"" ws

number ::= ("-"? ([0-9] | [1-9] [0-9]*)) ("." [0-9]+)? ([eE] [-+]? [0-9]+)? ws

ws ::= ([ \t\n] ws)?
"""




"""Models is a singleton class that centralizes model configuration and creation."""

from typing import Any, Optional, Union, Callable

import os, json, re
from copy import copy

from pprint import pformat

import logging
logger = logging.getLogger(__name__)

from .model import (
    GenConf,
    Model
)

from .utils import (
    dict_merge,
    expand_path
)




class Models:
    """Model and template format directory that unifies (and simplifies) model access and configuration.

    This env variable is checked and used during initialization:
        SIBILA_MODELS: ';'-delimited list of folders where to find: models.json, formats.json and model files.


    = Model Directory =

    Useful to create models from resource names like "llamacpp:openchat" or "openai:gpt-4". 
    This makes it simple to change a model, store model settings, to compare model outputs, etc.
    
    User can add new entries from script or with JSON filenames, via the add() call.
    New directory entries with the same name are merged into existing ones for each added config.

    Uses file base_models.json in this script's directory for the initial defaults, 
    which the user can augment by calling setup() with own config files or directly adding model config with set_model().

    An example of a model directory JSON config file:

    ``` json
    {
        # "llamacpp" is a provider, you can then create models with names 
        # like "provider:model_name", for ex: "llamacpp:openchat"
        "llamacpp": { 

            "_default": { # place here default args for all llamacpp: models.
                "genconf": {"temperature": 0.0}
                # each model entry below can then override as needed
            },
            
            "openchat": { # a model definition
                "name": "openchat-3.5-1210.Q4_K_M.gguf",
                "format": "openchat" # chat template format used by this model
            },

            "phi2": {
                "name": "phi-2.Q4_K_M.gguf", # model filename
                "format": "phi2",
                "genconf": {"temperature": 2.0} # a hot-headed model
            },

            "oc": "openchat" 
            # this is a link: "oc" forwards to the "openchat" entry
        },

        # The "openai" provider. A model can be created with name: "openai:gpt-4"
        "openai": { 

            "_default": {}, # default settings for all OpenAI models
            
            "gpt-3.5": {
                "name": "gpt-3.5-turbo-1106" # OpenAI's model name
            },

            "gpt-4": {
                "name": "gpt-4-1106-preview"
            },
        },

        # "alias" entry is not a provider but a way to have simpler alias names.
        # For example you can use "alias:develop" or even simpler, just "develop" to create the model:
        "alias": { 
            "develop": "llamacpp:openchat",
            "production": "openai:gpt-3.5"
        }
    }
    ```

    = Format Directory =

    Detects chat templates from model name/filename or uses from metadata if possible.

    This directory can be setup from a JSON file or by calling set_format().

    Any new entries with the same name replace previous ones on each new call.
    
    Initializes from file base_formats.json in this module's directory.


    Example of a "formats.json" file:

    ``` json
    {
        "chatml": {
            # template is a Jinja template for this model
            "template": "{% for message in messages %}..."
        },

        "openchat": {
            "match": "openchat", # a regexp to match model name or filename
            "template": "{{ bos_token }}..."
        },    

        "phi": {
            "match": "phi",
            "template": "..."
        },

        "phi2": "phi",
        # this is a link: "phi2" -> "phi"
    }
    ```

    Jinja2 templates receive a standard ChatML messages list (created from a Thread) and must deal with the following:

    - In models that don't use a system message, template must take care of prepending it to first user message.
    
    - The add_generation_prompt template variable is always set as True.

    """


    # ======================================================== Model directory
    models_dir: Any = None # model directory configuration => Union[dict[str,Any],None]
    """Model directory dict."""

    models_search_path: list[str] = [] 
    """Model search path: list of folders with models."""

    genconf: Union[GenConf,None] = None
    """Default GenConf for created models."""

    ENV_VAR_NAME = "SIBILA_MODELS"

    PROVIDER_CONF = {
        "llamacpp": {
            "mandatory": ["name"],
            "flags": ["name_passthrough"]
        },
        "openai": {
            "mandatory": ["name"],
            "flags": ["name_passthrough"]
        }
    }
    ALL_PROVIDER_NAMES = list(PROVIDER_CONF.keys()) + ["alias"] # providers + "alias"

    EMPTY_MODELS_DIR: dict = {
        "llamacpp": {},
        "openai": {},
        "alias": {}
    }

    MODELS_CONF_FILENAME = "models.json"
    MODELS_BASE_CONF_FILENAME = "base_" + MODELS_CONF_FILENAME

    DEFAULT_ENTRY_NAME = "_default"


    # ======================================================== Format directory    
    formats_dir: Any = None # model directory configuration => Union[dict[str,Any],None]
    """Format directory dict."""

    EMPTY_FORMATS_DIR: dict = {}

    FORMATS_CONF_FILENAME = "formats.json"
    FORMATS_BASE_CONF_FILENAME = "base_" + FORMATS_CONF_FILENAME











    @classmethod
    def setup(cls,
              path: Optional[Union[str,list[str]]] = None,
              clear: bool = False,
              add_cwd: bool = True,
              load_base: bool = True,
              load_from_env: bool = True):
        """Initialize models and formats directory from given model files folder and/or contained configuration files.
        Path can start with "~/" current account's home directory.

        Args:
            path: Path to a folder or to "models.json" or "formats.json" configuration files. Defaults to None which tries to initialize from defaults and env variable.
            clear: Set to clear existing directories before loading from path arg.
            add_cwd: Add current working directory to search path.
            load_base: Whether to load "base_models.json" and "base_formats.json".
            load_from_env: Load from SIBILA_MODELS env variable?
        """
        
        if clear:
            cls.clear()

        cls._ensure(add_cwd, 
                    load_base,
                    load_from_env)

        if path is not None:
            if isinstance(path, str):
                path_list = [path]
            else:
                path_list = path

            cls._read_any(path_list)
        
    


    @classmethod
    def clear(cls):
        """Clear directories. Member genconf is not cleared."""
        cls.models_dir = None
        cls.models_search_path = []
        cls.formats_dir = None





    
    @classmethod
    def info(cls,
             verbose: bool = False) -> str:
        """Return information about current setup.

        Args:
            verbose: If False, formats directory values are abbreviated. Defaults to False.

        Returns:
            Textual information about the current setup.
        """
        
        cls._ensure()

        out = ""
        
        out += f"Models search path: {cls.models_search_path}\n"
        out += f"Models directory:\n{pformat(cls.models_dir, sort_dicts=False)}\n"
        out += f"Model Genconf:\n{cls.genconf}\n"

        if not verbose:
            fordir = {}
            for key in cls.formats_dir:
                fordir[key] = copy(cls.formats_dir[key])
                if isinstance(fordir[key], dict) and "template" in fordir[key]:
                    fordir[key]["template"] = fordir[key]["template"][:14] + "..."
        else:
            fordir = cls.formats_dir

        out += f"Formats directory:\n{pformat(fordir)}"

        return out
    

    




    # ================================================================== Models
    
    @classmethod
    def create(cls,
               res_name: str,

               # common to all providers
               genconf: Optional[GenConf] = None,
               ctx_len: Optional[int] = None,

               # model-specific overriding:
               **over_args: Union[Any]) -> Model:
        """Create a model.

        Args:
            res_name: Resource name in the format: provider:model_name, for example "llamacpp:openchat".
            genconf: Optional model generation configuration. Overrides set_genconf() value and any directory defaults. Defaults to None.
            ctx_len: Maximum context length to be used. Overrides directory defaults. Defaults to None.
            over_args: Model-specific creation args, which will override default args set in model directory.

        Returns:
            Model: the initialized model.
        """
               
        cls._ensure()        
            
        # resolve "alias:name" res names, or "name": "link_name" links
        provider,name = cls.resolve_model_urn(res_name)
        # arriving here, prov as a non-link dict entry
        logger.debug(f"Resolved model '{res_name}' to '{provider}','{name}'")

        prov = cls.models_dir[provider]
        
        if name in prov:
            model_args = prov[name]
    
            # _default(if any) <- model_args <- over_args
            args = (prov.get(cls.DEFAULT_ENTRY_NAME)).copy() or {}
            args.update(model_args)        
            args.update(over_args)
    
        else:                
            prov_conf = cls.PROVIDER_CONF[provider]    

            if "name_passthrough" in prov_conf["flags"]:
                model_args = {
                    "name": name                
                }
            else:
                raise ValueError(f"Model '{name}' not found in provider '{provider}'")
            
            args = {}
            args.update(model_args)
            args.update(over_args)


        # override genconf, ctx_len
        if genconf is None:
            genconf = cls.genconf

        if genconf is not None:
            args["genconf"] = genconf

        elif "genconf" in args and isinstance(args["genconf"], dict):
            # transform dict into a GenConf instance:
            args["genconf"] = GenConf.from_dict(args["genconf"])

        if ctx_len is not None:
            args["ctx_len"] = ctx_len

        logger.debug(f"Creating model '{provider}:{name}' with resolved args: {args}")


        model: Model
        if provider == "llamacpp":

            # resolve filename -> path
            path = cls._locate_file(args["name"])
            if path is None:
                raise FileNotFoundError(f"File not found in '{res_name}' while looking for file '{args['name']}'. Make sure you called Models.setup() with a path to the file's folder")

            logger.debug(f"Resolved llamacpp model '{args['name']}' to '{path}'")
            
            del args["name"]
            args["path"] = path
                        
            from .llamacpp import LlamaCppModel

            model = LlamaCppModel(**args)

        
        elif provider == "openai":

            from .openai import OpenAIModel
                    
            model = OpenAIModel(**args)
            
        """
        elif provider == "hf":
            from .hf import HFModel
            
            model = HFModel(**args)
        """
           
        return model




    @classmethod
    def add_models_search_path(cls,
                               path: Union[str,list[str]]):
        """Prepends new paths to model files search path.

        Args:
            path: A path or list of paths to add to model search path.
        """

        cls._ensure()

        prepend_path(cls.models_search_path, path)

        logger.debug(f"Adding '{path}' to search_path")

         

    @classmethod
    def set_genconf(cls,
                    genconf: GenConf):
        """Set the GenConf to use as default for model creation.

        Args:
            genconf: Model generation configuration.
        """
        cls.genconf = genconf



    @classmethod
    def list_models(cls,
                    name_query: str,
                    providers: list[str],
                    resolved_values: bool) -> dict:
        """List format entries matching query.

        Args:
            name_query: Case-insensitive substring to match model names. Empty string for all.
            providers: Filter by these exact provider names. Empty list for all.
            resolved_values: Return resolved entries or raw ones.

        Returns:
            A dict where keys are model res_names and values are respective entries.
        """

        cls._ensure()

        out = {}

        name_query = name_query.lower()

        for prov_name in cls.models_dir:

            if providers and prov_name not in providers:
                continue

            prov_dic = cls.models_dir[prov_name]

            for name in prov_dic:

                if name == cls.DEFAULT_ENTRY_NAME:
                    continue

                if name_query and name_query not in name.lower():
                    continue

                entry_res_name = prov_name + ":" + name

                if resolved_values:
                    res = cls.get_model_entry(entry_res_name) # type: ignore[assignment]
                    if res is None:
                        continue
                    else:
                        val = res[1]
                else:
                    val = prov_dic[name]

                out[entry_res_name] = val

        return out





    @classmethod
    def get_model_entry(cls,
                        res_name: str) -> Union[tuple[str,dict],None]:
        """Get a resolved model entry. Resolved means following any links.

        Args:
            res_name: Resource name in the format: provider:model_name, for example "llamacpp:openchat".

        Returns:
            Resolved entry (res_name,dict) or None if not found.
        """
               
        cls._ensure()        
            
        # resolve "alias:name" res names, or "name": "link_name" links
        provider,name = cls.resolve_model_urn(res_name)
        # arriving here, prov as a non-link dict entry
        logger.debug(f"Resolved model '{res_name}' to '{provider}','{name}'")

        prov = cls.models_dir[provider]
        
        if name in prov:
            return provider + ":" + name, prov[name]
        else:
            return None

    @classmethod
    def has_model_entry(cls,
                        res_name: str) -> bool:
        return cls.get_model_entry(res_name) is not None




    @classmethod
    def set_model(cls,
                  res_name: str,
                  model_name: str,
                  format_name: Optional[str] = None,
                  genconf: Optional[GenConf] = None):
        """Add model configuration for given res_name.

        Args:
            res_name: A name in the form "provider:model_name", for example "openai:gtp-4".
            model_name: Model name or filename identifier.
            format_name: Format name used by model. Defaults to None.
            genconf: Base GenConf to use when creating model. Defaults to None.

        Raises:
            ValueError: If unknown provider.
        """

        cls._ensure()
        
        provider,name = provider_name_from_urn(res_name, False)
        if provider not in cls.ALL_PROVIDER_NAMES:
            raise ValueError(f"Unknown provider '{provider}' in '{res_name}'")
        
        entry: dict = {
            "name": model_name
        }

        if format_name:
            if not cls.has_format_entry(format_name):
                raise ValueError(f"Could not find format '{format_name}'")
            entry["format"] = format_name

        if genconf:
            entry["genconf"] = genconf.as_dict()

        cls.models_dir[provider][name] = entry

       


    @classmethod
    def update_model(cls,
                     res_name: str,
                     model_name: Optional[str] = None,
                     format_name: Optional[str] = None,
                     genconf: Union[GenConf,str,None] = None):
        
        """update model fields

        Args:
            res_name: A name in the form "provider:model_name", for example "openai:gtp-4".
            model_name: Model name or filename identifier. Defaults to None.
            format_name: Format name used by model. Use "" to delete. Defaults to None.
            genconf: Base GenConf to use when creating model. Defaults to None.

        Raises:
            ValueError: If unknown provider.
        """

        cls._ensure()
        
        provider,name = provider_name_from_urn(res_name, False)
        if provider not in cls.ALL_PROVIDER_NAMES:
            raise ValueError(f"Unknown provider '{provider}' in '{res_name}'")
        
        entry = cls.models_dir[provider][name]

        if model_name:
            entry["name"] = model_name

        if format_name is not None:
            if format_name != "":
                if not cls.has_format_entry(format_name):
                    raise ValueError(f"Could not find format '{format_name}'")
                entry["format"] = format_name
            else:
                del entry["format"]

        if genconf is not None:
            if genconf != "":
                entry["genconf"] = genconf
            else:
                del entry["genconf"]




    @classmethod
    def set_model_link(cls,
                       res_name: str,
                       link_name: str):
        """Create a model link into another model.

        Args:
            res_name: A name in the form "provider:model_name", for example "openai:gtp-4".
            link_name: Name of model this entry links to.

        Raises:
            ValueError: If unknown provider.
        """
        
        cls._ensure()
        
        provider,name = provider_name_from_urn(res_name, True)
        if provider not in cls.ALL_PROVIDER_NAMES:
            raise ValueError(f"Unknown provider '{provider}' in '{res_name}'")
        
        # first: ensure link_name is a res_name
        if ':' not in link_name:
            link_name = provider + ":" + link_name

        if not cls.has_model_entry(link_name):
            raise ValueError(f"Could not find linked model '{link_name}'")

        # second: check link name is without provider if same
        link_split = link_name.split(":")
        if len(link_split) == 2:
            if link_split[0] == provider: # remove same "provider:"
                link_name = link_split[1]

        cls.models_dir[provider][name] = link_name




    @classmethod
    def delete_model(cls,
                     res_name: str):
        """Delete a model entry.

        Args:
            res_name: Model entry in the form "provider:name".
        """

        cls._ensure()

        provider,name = cls.resolve_model_urn(res_name)

        prov = cls.models_dir[provider]

        del prov[name]






    @classmethod
    def resolve_model_urn(cls,
                          res_name: str) -> tuple[str,str]:
        """
        Checks format and if provider exists, follows string links until a dict or non-existent name key.
        res_name must be in format provider:model_name
        Returns tuple of provider_name, model_name. provider_name must exist, model_name may not
        """

        while True:
            provider, name = provider_name_from_urn(res_name, True)
                
            if provider not in cls.ALL_PROVIDER_NAMES:
                raise ValueError(f"Don't know how to handle provider '{provider}'. Can only handle the following providers: {cls.ALL_PROVIDER_NAMES}")

            prov = cls.models_dir[provider]

            if name in prov and isinstance(prov[name], str): # follow string link
                res_name = prov[name]
                if ":" not in res_name: # a local provider link
                    res_name = provider + ":" + res_name
                    
            elif provider == "alias" and name not in prov: # no alias with that name
                raise ValueError(f"Alias not found for '{name}'. Did you mean 'llamacpp:{name}' or 'openai:{name}'?")
                
            else: 
                break
            
        return provider, name
    

    @classmethod
    def save_models(cls,
                    path: Optional[str] = None):
        if path is None:
            if len(cls.models_search_path) != 1:
                raise ValueError("No path arg provided and multiple path in cls.search_path. Don't know where to save.")
            
            path = os.path.join(cls.models_search_path[0], "models.json")

        with open(path, "w", encoding="utf-8") as f:
            json.dump(cls.models_dir, f, indent=4)

        return path







    # =========================================================================== Formats

    @classmethod
    def list_formats(cls,
                     name_query: str,
                     resolved_values: bool) -> dict:
        """List format entries matching query.

        Args:
            name_query: Case-insensitive substring to match format names. Empty string for all.
            resolved_values: Return resolved entries or raw ones.

        Returns:
            A dict where keys are format names and values are respective entries.
        """

        cls._ensure()

        out = {}

        name_query = name_query.lower()

        for name in cls.formats_dir.keys():

            if name_query and name_query not in name.lower():
                continue

            val = cls.formats_dir[name]

            if resolved_values:
                res = cls.get_format_entry(name)
                if res is None:
                    continue
                else:
                    val = res[1]

            out[name] = val

        return out




    @classmethod
    def get_format_entry(cls,
                         name: str) -> Union[tuple[str,dict],None]:
        """Get a resolved format entry by name, following links if required.

        Args:
            name: Format name.

        Returns:
            Tuple of (resolved_name, format_entry).
        """

        cls._ensure()

        na = name.lower()
        while na in cls.formats_dir.keys():
            val = cls.formats_dir[na]
            if isinstance(val, str): # str means link -> follow it
                na = val
            else:
                logger.debug(f"Format get('{name}'): found '{na}' entry")
                return na, resolve_format_entry(cls.formats_dir,
                                                na,
                                                val)
        return None
    
    @classmethod
    def has_format_entry(cls,
                         name: str) -> bool:
        return cls.get_format_entry(name) is not None



    @classmethod
    def get_format_template(cls,
                            name: str) -> Union[str,None]:
        """Get a format template by name, following links if required.

        Args:
            name: Format name.

        Returns:
            Resolved format template str.
        """

        res = cls.get_format_entry(name)
        return None if res is None else res[1]["template"]




    @classmethod
    def match_format_entry(cls,
                           name: str) -> Union[tuple[str,dict],None]:
        """Search the formats registry, based on model name or filename.

        Args:
            name: Name or filename of model.

        Returns:
            Tuple (name, format_entry) where name is a resolved name. Or None if none found.
        """

        cls._ensure()

        return search_format(cls.formats_dir, name)


    @classmethod
    def match_format_template(cls,
                              name: str) -> Union[str,None]:
        """Search the formats registry, based on model name or filename.

        Args:
            name: Name or filename of model.

        Returns:
            Format template or None if none found.
        """

        res = cls.match_format_entry(name)

        return None if res is None else res[1]["template"]






    @staticmethod
    def folder_match_format_entry(model_path: str) -> Union[tuple[str,dict],None]:
        """Locally search for format in a formats.json file located in the same folder as the model.
        Doesn't add format to class formats directory.

        Args:
            model_path: Model path.

        Returns:
            Tuple (name, format_entry) where name is a resolved name. Or None if none found.
        """

        folder_path = os.path.dirname(model_path)
        formats_path = os.path.join(folder_path, Models.FORMATS_CONF_FILENAME)
        if not os.path.isfile(formats_path):
            return None

        logger.info(f"Loading local formats conf from '{formats_path}'")

        formats_dir:dict = {}
        try:
            update_dir_json(formats_dir, formats_path)
        except Exception:
            raise ValueError(f"Could not load 'formats.json' at '{formats_path}', while looking for model format. "
                             "Please verify that he JSON syntax is correct.")

        sanity_check_formats(formats_dir)

        model_id = os.path.basename(model_path)
        return search_format(formats_dir, model_id)


    @staticmethod
    def folder_match_format_template(model_path: str) -> Union[str,None]:
        """Locally search for format in a formats.json file located in the same folder as the model.
        Doesn't add format to class formats directory.

        Args:
            model_path: Model path.

        Returns:
            Format template.
        """

        res = Models.folder_match_format_entry(model_path)
        return None if res is None else res[1]["template"]




    @classmethod
    def is_format_supported(cls,
                            model_id: str) -> bool:
        """Checks if there's template support for a model with this name.

        Args:
            model_id: Model filename or general name.

        Returns:
            True if Models knows the format.
        """

        return cls.match_format_entry(model_id) is not None







    @classmethod
    def set_format(cls,
                   name: str,
                   match: str,
                   template: str):
        """Add a format entry to the format directory.

        Args:
            name: Format entry name.
            match: Regex that matches names/filenames that use this format.
            template: The Chat template format in Jinja2 format
        """

        cls._ensure()

        if "{{" not in template: # a link_name for the template
            if not cls.has_format_entry(template):
                raise ValueError(f"Could not find linked template entry '{template}'.")

        entry = {
            "match": match,
            "template": template
        }
        cls.formats_dir[name] = entry        
        



    @classmethod
    def set_format_link(cls,
                        name: str,
                        link_name: str):
        """Add a format link entry to the format directory.

        Args:
            name: Format entry name.
            link_name: Name of format that this entry links to.
        """

        cls._ensure()

        if not cls.has_format_entry(link_name):
            raise ValueError(f"Could not find linked entry '{link_name}'.")

        cls.formats_dir[name] = link_name
        



    @classmethod
    def delete_format(cls,
                      name: str):
        """Delete a format entry.

        Args:
            name: Format entry name.
        """

        cls._ensure()

        if not cls.has_format_entry(name):
            raise ValueError(f"Format name '{name}' not found.")
        
        del cls.formats_dir[name]





    @classmethod
    def merge_from(cls,
                   path: str,
                   preserve_current: bool = True):
        path = expand_path(path)

        if preserve_current:
            with open(path, "r", encoding="utf-8") as f:
                new_dir = json.load(f)
            new_dir.update(cls.formats_dir)
            cls.formats_dir = new_dir

        else: # normal update: new with the same name will override current
            update_dir_json(cls.formats_dir, path)

        sanity_check_formats(cls.formats_dir)


    @classmethod
    def save_formats(cls,
                     path: Optional[str] = None):
        if path is None:
            if len(cls.models_search_path) != 1:
                raise ValueError("No path arg provided and multiple path in cls.search_path. Don't know where to save.")
            
            path = os.path.join(cls.models_search_path[0], "formats.json")

        with open(path, "w", encoding="utf-8") as f:
            json.dump(cls.formats_dir, f, indent=4)

        return path






    # ================================================================== Lower level

    @classmethod
    def _ensure(cls,
                add_cwd: bool = True,
                load_base: bool = True,
                load_from_env: bool = True):
        """Make sure class is initialized.

        Env variable checked:
            SIBILA_MODELS: ';'-delimited folder list where to find models.        
        """

        if cls.models_dir is not None: 
            return
            
        # model and format dirs
        cls.models_dir = cls.EMPTY_MODELS_DIR
        cls.formats_dir = cls.EMPTY_FORMATS_DIR

        if add_cwd:
            # add "." to search path, so that paths relative paths work
            cls.add_models_search_path(".")

        path: Union[str, None]

        if load_base:
            # read base models and formats in same folder as this script file
            base_path: str = os.path.abspath(__file__)
            base_path = os.path.dirname(base_path)

            # read base_models.json
            path = os.path.join(base_path, cls.MODELS_BASE_CONF_FILENAME)
            if os.path.isfile(path):
                cls._read_models(path,
                                 add_folder_to_search_path=False)

            # read base_formats.json
            path = os.path.join(base_path, cls.FORMATS_BASE_CONF_FILENAME)
            if os.path.isfile(path):
                cls._read_formats(path)

        # check env var
        if load_from_env:
            path = os.environ.get(cls.ENV_VAR_NAME)
            if path is not None:
                path_list = path.split(";")
                cls._read_any(path_list)





    @classmethod
    def _read_models(cls,
                     models_path: str,
                     add_folder_to_search_path: bool = True):
 
        logger.info(f"Loading models conf from: '{models_path}'")
        merge_dir_json(cls.models_dir,
                       models_path)

        cls._sanity_check_models()

        if add_folder_to_search_path:
            dir_path = os.path.dirname(models_path)
            cls.add_models_search_path(dir_path)


    @classmethod
    def _read_formats(cls,
                      formats_path: str):
        
        logger.info(f"Loading formats conf from '{formats_path}'")
        update_dir_json(cls.formats_dir, formats_path)

        sanity_check_formats(cls.formats_dir)
        


    @classmethod
    def _read_folder(cls,
                     path: str):

        if not os.path.isdir(path):
            raise OSError(f"Directory not found: '{path}'")
        
        cls.add_models_search_path(path)

        # read models
        models_path = os.path.join(path, cls.MODELS_CONF_FILENAME)
        if os.path.isfile(models_path):
            cls._read_models(models_path)

        # read formats
        formats_path = os.path.join(path, cls.FORMATS_CONF_FILENAME)
        if os.path.isfile(formats_path):
            cls._read_formats(formats_path)
           


    @classmethod
    def _read_any(cls,
                  path: Union[str,list[str]]):
        
        if isinstance(path, str):
            path_list = [path]
        else:
            path_list = path
        
        for path in path_list:
            path = expand_path(path)

            if os.path.isdir(path):
                cls._read_folder(path)

            elif path.find(cls.MODELS_CONF_FILENAME) >= 0:
                cls._read_models(path)

            elif path.find(cls.FORMATS_CONF_FILENAME) >= 0:
                cls._read_formats(path)



    @classmethod
    def _locate_file(cls,
                     path: str) -> Union[str,None]:
        
        if os.path.isabs(path): # absolute?
            if os.path.isfile(path):
                return path
            else:
                return None
        
        for dir in cls.models_search_path:
            full_path = os.path.join(dir, path)
            if os.path.isfile(full_path):
                return full_path
            
        return None

        


    @classmethod
    def _sanity_check_models(cls):
        
        for prov in cls.models_dir.keys():
            if prov == "alias":
                for name,link_name in cls.models_dir[prov].items():
                    if not isinstance(link_name, str):
                        raise ValueError(f"Alias entries must be strings at alias:{name}")
                        
            else: # real providers
                if prov not in cls.PROVIDER_CONF.keys():
                    raise ValueError(f"Don't know how to handle provider '{prov}'. Can only handle the following providers: {cls.PROVIDER_CONF.keys()}")
    
                # check if mandatory keys are in each model entry
                mandatory_keys = cls.PROVIDER_CONF[prov]["mandatory"]
                prov_models = cls.models_dir[prov]
                for model_name in prov_models.keys():
                    if model_name == cls.DEFAULT_ENTRY_NAME or isinstance(prov_models[model_name], str):
                        continue # skip "_default" args entry and string links
                    
                    model_entry = prov_models[model_name]
                    if not all(mand in model_entry for mand in mandatory_keys):
                        raise ValueError(f"Models entry '{prov}:{model_name}' doesn't have all mandatory keys for this provider ({mandatory_keys})") 
                               
                
        # ensure all providers have their own entry
        for p in cls.PROVIDER_CONF.keys():
            if p not in cls.models_dir.keys():
                cls.models_dir[p] = {}

        if "alias" not in cls.models_dir:
            cls.models_dir["alias"] = {}
            





# ======================================================================== Utils


def prepend_path(base_list: list[str],
                 path: Union[str,list[str]]):
    
    if isinstance(path, str):
        path_list = [path]
    else:
        path_list = path.copy()

    for p in path_list[::-1]:
        p = expand_path(p)
        while p in base_list:
            base_list.remove(p)            
        base_list.insert(0,p)




def merge_dir_json(dir: dict,
                   path: str):
    
    path = os.path.abspath(path)
    
    with open(path, "r", encoding="utf-8") as f:
        in_dir = json.load(f)

    dict_merge(dir, in_dir)



    



def provider_name_from_urn(res_name: str,
                           allow_alias_provider: bool) -> tuple[str,str]:
    if ":" in res_name:
        provider_name = tuple(res_name.split(":"))
        if len(provider_name) > 2:
            raise ValueError(f"Model resource name must be in the format provider:model_name (for '{res_name}')")
    else:
        if allow_alias_provider:
            provider_name = "alias", res_name # type: ignore[assignment]
        else:
            raise ValueError(f"Alias not allowed (for '{res_name}')")
        
    return provider_name # type: ignore[return-value]





def update_dir_json(dir: dict,
                    path: str):
    path = os.path.abspath(path)
    with open(path, "r", encoding="utf-8") as f:
        new_dir = json.load(f)
        dir.update(new_dir)



def resolve_format_entry(formats_dir: dict,
                         name: str,
                         val: Union[dict,str]):
    val = copy(val)
    
    if "{{" not in val["template"]: # type: ignore[index] # a link to another template entry
        linked_name = val["template"] # type: ignore[index]
        if linked_name not in formats_dir:
            raise ValueError(f"Broken template link at '{name}': '{linked_name}' does not exist")
        val2 = formats_dir[linked_name]
        val["template"] = val2["template"] # type: ignore[index]
        
    return val




def search_format(formats_dir: dict,
                  model_id: str) -> Union[tuple[str,dict],None]:

    # TODO: cache compiled re patterns in "_re" entries

    for name,val in formats_dir.items():
        if isinstance(val, str): # a link: ignore when searching
            continue
        if "match" not in val:
            continue
            
        patterns = val["match"]
        if isinstance(patterns, str):
            patterns = [patterns]
            
        for pat in patterns:
            if re.search(pat, model_id, flags=re.IGNORECASE):
                logger.debug(f"Format search for '{model_id}' found '{name}' entry")
                return name, resolve_format_entry(formats_dir,
                                                  name,
                                                  val)
                            
    return None


def sanity_check_formats(formats_dir: dict):

    # sanity check complete directory
    for name,val in formats_dir.items():
        if isinstance(val, str): # a link -> does pointed-to exist?
            if val not in formats_dir.keys():
                raise ValueError(f"Entry '{name}' points to non-existent entry '{val}'")
        else:
            if not isinstance(val, dict):
                raise ValueError(f"Entry '{name}' must be a dict")
                
            if "template" not in val:
                raise ValueError(f"Entry '{name}' must have a 'template' value")
                    








"""Threads of messages are the way to communicate with models. 
For a quick extraction or classification, a thread with a simple IN message and possibly model instructions can be used.
For more sophisticated queries that depend on previous interactions, IN (user or input) and OUT (model response) messages can be added.

- Msg: A message of type IN (user query), OUT (assistant answer) or SYS (initial assistant instructions).
- Thread: A sequence of Msgs alternating between user and model.
"""

from typing import Any, Optional, Union, Callable
from typing_extensions import Self
from enum import IntEnum
from collections.abc import Sequence

from copy import copy, deepcopy
import json

import logging
logger = logging.getLogger(__name__)




class MsgKind(IntEnum):
    """Enumeration for kinds of messages in a Thread."""

    IN = 0
    """Input message, from user."""

    OUT = 1
    """Model output message."""

    INST = 2
    """Initial instructions for model."""

    @staticmethod
    def kind_from_chatml_role(role: str) -> Any: # Any=MsgKind
        KIND_FROM_CHATML: dict = {"user": MsgKind.IN, "assistant": MsgKind.OUT, "system": MsgKind.INST}
        kind = KIND_FROM_CHATML.get(role)
        if kind is None:
            raise ValueError(f"Unknown ChatML role '{role}'.")
        else:
            return kind

    @staticmethod
    def chatml_role_from_kind(kind: Any) -> str: # Any=MsgKind
        CHATML_FROM_KIND: dict = {MsgKind.IN: "user", MsgKind.OUT: "assistant", MsgKind.INST: "system"}
        return CHATML_FROM_KIND.get(kind) # type: ignore[return-value]




class Thread(Sequence):
    """A sequence of messages alternating between IN ("user" role) and OUT ("assistant" role).

    Stores a special initial INST information (known as "system" role in ChatML) providing instructions to the model.
    Some models don't use system instructions - in those cases it's prepended to first IN message.

    Messages are kept in a strict IN,OUT,IN,OUT,... order. To enforce this, if two IN messages are added, the second just appends to the text of the first.
    
    Attributes:
        inst: Text for system instructions.
    """

    inst: str
    """Text for system instructions, defaults to empty string"""

    join_sep: str
    """Separator used when message text needs to be joined. Defaults to '\\n'"""

    _msgs: list[str]
    """List of thread messages"""
    

    def __init__(self,
                 t: Optional[Union[Any,list,str,dict,tuple]] = None, # Any=Thread
                 inst: str = '',
                 join_sep: str = "\n"):
        """
        Examples:
            Creation with a list of  messages

            >>> from sibila import Thread, MsgKind
            >>> th = Thread([(MsgKind.IN, "Hello model!"), (MsgKind.OUT, "Hello there human!")],
            ...             inst="Be helpful.")
            >>> print(th)
            inst=█Be helpful.█, sep='\\n', len=2
            0: IN=█Hello model!█
            1: OUT=█Hello there human!█

            Adding messages

            >>> from sibila import Thread, MsgKind
            >>> th = Thread(inst="Be helpful.")
            >>> th.add(MsgKind.IN, "Can you teach me how to cook?")
            >>> th.add_IN("I mean really cook as a chef?") # gets appended
            >>> print(th)
            inst=█Be helpful.█, sep='\\n', len=1
            0: IN=█Can you teach me how to cook?\\nI mean really cook as a chef?█

            Another way to add a message

            >>> from sibila import Thread, MsgKind
            >>> th = Thread(inst="Be informative.")
            >>> th.add_IN("Tell me about kangaroos, please?")
            >>> th += "They are so impressive." # appends text to last message
            >>> print(th)
            inst=█Be informative.█, sep='\\n', len=1
            0: IN=█Tell me about kangaroos, please?\\nThey are so impressive.█
            
            Return thread as a ChatML message list

            >>> from sibila import Thread, MsgKind
            >>> th = Thread([(MsgKind.IN, "Hello model!"), (MsgKind.OUT, "Hello there human!")], 
            ...             inst="Be helpful.")
            >>> th.as_chatml()
            [{'role': 'system', 'content': 'Be helpful.'},
             {'role': 'user', 'content': 'Hello model!'},
             {'role': 'assistant', 'content': 'Hello there human!'}]

        Args:
            t: Can initialize from a Thread, from a list (containing messages in any format accepted in _parse_msg()) or a single message as an str, an (MsgKind,text) tuple or a dict. Defaults to None.
            join_sep: Separator used when message text needs to be joined. Defaults to "\\n".

        Raises:
            TypeError: On invalid args passed.
        """

        if isinstance(t, Thread):
            self._msgs = t._msgs.copy()
            self.inst = t.inst
            self.join_sep = t.join_sep
        else:
            self._msgs = []
            self.inst = inst
            self.join_sep = join_sep

            if t is not None:
                self.concat(t)



    def clear(self):
        """Delete all messages and clear inst."""
        self.inst = ""
        self._msgs = []
            
    

    @property
    def last_kind(self) -> MsgKind:
        """Get kind of last message in thread .

        Returns:
            Kind of last message or MsgKind.IN if empty.
        """
        length = len(self._msgs)
        if not length: # empty: assume IN
            return MsgKind.IN
        else:
            return Thread._kind_from_pos(length - 1)


    @property
    def last_text(self) -> str:
        """Get text of last message in thread .

        Returns:
            Last message text.

        Raises:
            IndexError: If thread is empty.
        """
        length = len(self._msgs)
        if not length: # empty
            raise IndexError("Thread is empty")
        else:
            return self._msgs[-1]


    def add(self, 
            t: Union[str,tuple,dict,MsgKind],
            text: Optional[str] = None):
        """Add a message to Thread by parsing a mix of types.

        Accepts any of these argument combinations:

        - t=MsgKind, text=str
        - t=str, text=None -> uses last thread message's MsgKind
        - (MsgKind, text)
        - {"kind": "...", text: "..."}
        - {"role": "...", content: "..."} - ChatML format

        Args:
            t: One of the accepted types listed above.
            text: Message text if first type is MsgKind. Defaults to None.
        """
        
        kind, text = self._parse_msg(t, text)

        if kind == MsgKind.INST:
            self.inst = self.join_text(self.inst, text)
        else:
            if kind == self.last_kind and len(self._msgs):
                self._msgs[-1] = self.join_text(self._msgs[-1], text)
            else:
                self._msgs.append(text) # in new kind


            
    def addx(self, 
             path: Optional[str] = None, 
             text: Optional[str] = None,
             kind: Optional[MsgKind] = None):
        """Add message with text from a supplied arg or loaded from a path.

        Args:
            path: If given, text is loaded from an UTF-8 file in this path. Defaults to None.
            text: If given, text is added. Defaults to None.
            kind: MsgKind of message. If not given or the same as last thread message, it's appended to it. Defaults to None.
        """

        assert (path is not None) ^ (text is not None), "Only one of path or text"

        if path is not None:
            with open(path, 'r', encoding="utf-8") as f:
                text = f.read()

        if kind is None: # use last message role, so that it gets appended
            kind = self.last_kind

        self.add(kind, text)



    def get_text(self,
                 index: int) -> str:
        """Return text for message at index.

        Args:
            index: Message index. Use -1 to get inst value.

        Returns:
            Message text at index.
        """        
        if index == -1:
            return self.inst
        else:
            return self._msgs[index]

    def set_text(self,
                 index: int,
                 text: str):        
        """Set text for message at index.

        Args:
            index: Message index. Use -1 to set inst value.
            text: Text to replace in message at index.
        """
        if index == -1:
            self.inst = text
        else:
            self._msgs[index] = text
            

    def concat(self,
               t: Optional[Union[Self,list,str,dict,tuple]]):
        """Concatenate a Thread or list of messages to the current Thread.

        Take care that the other list starts with an IN message, therefore, 
        if last message in self is also an IN kind, their text will be joined as in add().

        Args:
            t: A Thread or a list of messages. Otherwise a single message as in add().

        Raises:
            TypeError: If bad arg types provided.
        """
        if isinstance(t, Thread):
            for msg in t:
                self.add(msg)
            self.inst = self.join_text(self.inst, t.inst)

        elif isinstance(t, list): # message list
            for msg in t:
                self.add(msg)

        elif isinstance(t, str) or isinstance(t, dict) or isinstance(t, tuple): # single message
            self.add(t)

        else:
            raise TypeError("Arg t must be: Thread --or-- list[messages] --or-- an str, tuple or dict single message.")
    



    
    def load(self,
             path: str):
        """Load this Thread from a JSON file.

        Args:
            path: Path of file to load.
        """

        with open(path, 'r', encoding='utf-8') as f:
            js = f.read()
        state = json.loads(js)

        self._msgs = state["_msgs"]
        self.inst = state["inst"]
        self.join_sep = state["join_sep"]

    
    def save(self,
             path: str):
        """Serialize this Thread to JSON.

        Args:
            path: Path of file to save into.
        """
 
        state = {"_msgs": self._msgs,
                 "inst": self.inst,
                 "join_sep": self.join_sep
                 }
    
        json_str = json.dumps(state, indent=2, default=vars)
    
        with open(path, 'w', encoding='utf-8') as f:
            f.write(json_str)

    


    
    # = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = for convenience:
    
    def init_INST_IN(self,
                     inst_text: str,
                     in_text: str):
        """Initialize Thread with instructions and an IN message.

        Args:
            inst_text: Instructions text.
            in_text: Text for IN message.
        """
        self.clear()
        self += self.make_INST_IN(inst_text, in_text)


    def add_IN(self,
               in_text: str):
        """Appends an IN message to Thread.

        Args:
            in_text: Text for IN message.
        """
        self.add(MsgKind.IN, in_text)
    
    def add_OUT(self,
                out_text: str):
        """Appends an OUT message to Thread.

        Args:
            in_text: Text for OUT message.
        """
        self.add(MsgKind.OUT, out_text)

    
    def add_OUT_IN(self,
                   out_text: str,
                   in_text: str):
        """Appends an OUT message followed by an IN message.

        Args:
            out_text: Text for OUT message.
            in_text: Text for IN message.
        """        
        self.add(MsgKind.OUT, out_text)
        self.add(MsgKind.IN, in_text)



    @staticmethod
    def make_INST_IN(inst_text: str,
                     in_text: str) -> Any: # Any=Thread
        """Return an initialized Thread with instructions and an IN message.

        Args:
            inst_text: Instructions text.
            in_text: Text for IN message.
        """

        thread = Thread([(MsgKind.INST, inst_text),
                         (MsgKind.IN, in_text)]
                        )        
        return thread

    @staticmethod
    def make_IN(in_text: str) -> Any: # Any=Thread
        """Return an initialized Thread with an IN message.

        Args:
            in_text: Text for IN message.
        """

        thread = Thread([(MsgKind.IN, in_text)])        
        return thread

    @staticmethod
    def make_OUT_IN(out_text: str,
                    in_text: str) -> Any: # Any=Thread
        """Return an initialized Thread with an OUT message followed by an IN message.

        Args:
            out_text: Text for OUT message.
            in_text: Text for IN message.
        """
        thread = Thread([(MsgKind.OUT, out_text),
                         (MsgKind.IN, in_text)]
                        )
        return thread

    @staticmethod
    def ensure(query: Union[str,Any],
               inst: Optional[str] = None) -> Any: # Any=Thread
        """Utility to return a Thread from either a passed Thread or an str used as an IN message.

        Args:
            query: Thread or an str with the text of a single IN message to use as model input.
            inst: Instruction message for model. Will override Thread's inst. Defaults to None.

        Raises:
            TypeError: Arg query must be of type Thread or str.

        Returns:
            Initialized Thread object.
        """
        
        if isinstance(query, str):
            if inst is None:
                return Thread.make_IN(query)
            else:
                return Thread.make_INST_IN(inst, query)
        elif isinstance(query, Thread):
            if inst is not None:
                query = Thread(query, inst) # a clone
            return query
        else:
            raise TypeError("Arg query must be of type Thread or str")



    def msg_as_chatml(self,
                      index: int) -> dict:
        """Returns message in a ChatML dict.

        Args:
            index: Index of the message to return.

        Returns:
            A ChatML dict with "role" and "content" keys.
        """

        kind = Thread._kind_from_pos(index)
        role = MsgKind.chatml_role_from_kind(kind)
        text = self._msgs[index] if index >= 0 else self.inst
        return {"role": role, "content": text}
    

    def as_chatml(self) -> list[dict]:
        """Returns Thread as a list of ChatML messages.

        Returns:
            A list of ChatML dict elements with "role" and "content" keys.
        """
        msgs = []

        for index,msg in enumerate(self._msgs):
            if index == 0 and self.inst:
                msgs.append(self.msg_as_chatml(-1))
            msgs.append(self.msg_as_chatml(index))
            
        return msgs


    def has_text_lower(self,
                       text_lower: str) -> bool:
        """Can the lowercase text be found in one of the messages?

        Args:
            text_lower: The lowercase text to search for in messages.

        Returns:
            True if such text was found.
        """
        for msg in self._msgs:
            if text_lower in msg.lower():
                return True
            
        return False        
    

    
    def clone(self):
        """Return a copy of current Thread.

        Returns:
            A copy of this Thread.
        """
        return Thread(self)



    # = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = lower level
    def _parse_msg(self,
                   t: Union[str,tuple,dict,MsgKind],
                   text: Optional[str] = None) -> tuple[MsgKind,str]:
        
        """Parses a mix of types into (MsgKind, text).
        Accepts arg types:
            t:MsgKind, text:str
            t:str, text=None -> returns self.last_kind(), text
            (MsgKind, text)
            {"kind": "...", text: "..."}
            {"role": "...", content: "..."} - ChatML format

        Args:
            t: one of the accepted types.

        Returns:
            Tuple of parsed MsgKind, text
        """
        
        if text is not None: # t must be a kind
            if not isinstance(text, str):
                raise TypeError("If arg text is given, it must be an str.")
            if not isinstance(t, MsgKind):
                raise TypeError("If arg text is given, arg k must be of type MsgKind.")
            return t,text

        elif isinstance(t, str):
            return self.last_kind, t

        elif isinstance(t, tuple):
            if len(t) != 2 or not isinstance(t[0], MsgKind) or not isinstance(t[1], str):
                raise TypeError("Tuple must hold (MsgKind, str).")
            return t[0], t[1]
        
        elif isinstance(t, dict):
            if "role" in t:
                kind = MsgKind.kind_from_chatml_role(t["role"])
                if "content" not in t:
                    raise TypeError("A dict in format ChatML must include a content key.")
                return kind, t["content"]
            
            elif "kind" in t:
                if "text" not in t:
                    raise TypeError("Dict with 'kind' key must also include a 'text' key.")
                return t["kind"], t["text"]

            else:
                raise TypeError("Unknown dict format.")
        else:
            raise TypeError("""Args must be: MsgKind,str --or str,text=None --or-- (MsgKind, text) --or-- {"kind": "...", text: "..."} --or--  {"role": "...", content: "..."}.""")



    @staticmethod                     
    def _kind_from_pos(pos: int) -> MsgKind:
        if pos == -1:
            return MsgKind.INST
        else:
            return MsgKind.OUT if pos % 2 else MsgKind.IN


    def join_text(self,
                  a: str,
                  b: str,
                  sep_count: int = 1) -> str:        
        if b:
            if a:
                if a[-sep_count] != self.join_sep:
                    a = a + (self.join_sep * sep_count) + b
                else:
                    a += b
            else:
                a = b
        return a

            
    def __len__(self):
        return len(self._msgs)


    def __getitem__(self, # type: ignore[override]
                    index: int) -> tuple[MsgKind,str]:
        return Thread._kind_from_pos(index), self._msgs[index]


    def __delitem__(self, 
                    index: int):
        
        if index + 1 == len(self._msgs): # last: just delete
            del self._msgs[index]

        elif index == 0:
            raise IndexError("Can only delete at index 0 if len=1.")
        
        else:
            # augment next of the same kind with previous (of the same kind) text
            self._msgs[index + 1] = self.join_text(self._msgs[index - 1], 
                                                   self._msgs[index + 1])
            del self._msgs[index] # delete requested
            del self._msgs[index - 1] # delete previous


    def __iter__(self):
        class MsgIter:
            def __init__(self, thread):
                self.thread = thread
                self.curr = -1

            def __iter__(self):
                return self
            
            def __next__(self):
                self.curr += 1
                if self.curr < len(self.thread):
                    return self.thread[self.curr]
                else:
                    raise StopIteration

        return MsgIter(self)

    def __reversed__(self):
        return reversed(self._msgs)


    def __add__(self,
                other: Union[Self,list,str,dict,tuple]) -> Self:

        out = self.clone()
        out.concat(other)

        if isinstance(other, Thread):
            self.inst = self.join_text(self.inst, other.inst)

        return out


    def __str__(self):
        inst = self.inst.replace('\n', '\\n')
        sep = self.join_sep.replace('\n', '\\n')
        out = f"inst=█{inst}█, sep='{sep}', len={len(self._msgs)}"
        for index,text in enumerate(self._msgs):
            text = text.replace("\n", "\\n")
            kind = Thread._kind_from_pos(index)
            out += f"\n{index}: {kind.name}=█{text}█"
        return out
